import contextlib
import sys

if sys.version_info[:2] >= (3, 8):
    # TODO: Import directly (no need for conditional) when `python_requires = >= 3.8`
    from importlib.metadata import PackageNotFoundError, version  # pragma: no cover
else:
    from importlib_metadata import PackageNotFoundError, version  # pragma: no cover

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = "olorenchemengine"
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError


_OPTIONAL_IMPORTS_FOR_OCE_ONLINE = [
    "sklearn",
    "torch",
    "torch_geometric",
    "torch_scatter",
    "torch_sparse",
    "torch_cluster",
    "rdkit",
    "sklearn.metrics",
    "scipy",
    "scipy.stats",
    "joblib",
    "sklearn.model_selection",
    "tqdm",
    "sklearn.preprocessing",
    "rdkit.Chem",
    "torch.nn",
    "torch_geometric.nn",
    "torch_geometric.data",
    "torch.optim",
    "torch_geometric.utils",
    "torch.nn.functional",
    "torch_geometric.nn.inits",
    "rdkit.DataStructs",
    "rdkit.DataStructs.cDataStructs",
    "rdkit.Chem.AtomPairs",
    "rdkit.Chem.AtomPairs.Sheridan",
    "rdkit.Chem.Pharm2D",
    "torch_geometric.loader",
    "torch_geometric.data.data",
    "torch.utils",
    "torch.utils.data",
    "torch.autograd",
    "rdkit.Chem.rdchem",
    "pytorch_lightning",
    "rdkit.Chem.Scaffolds",
    "PIL",
    "selfies",
    "sklearn.cross_decomposition",
    "sklearn.decomposition",
    "sklearn.neighbors",
    "sklearn.manifold",
    "hyperopt",
    "hyperopt.pyll",
]

_mocked_imports = []

for imp in _OPTIONAL_IMPORTS_FOR_OCE_ONLINE:
    try:
        __import__(imp)
    except ImportError:
        import sys
        from unittest.mock import MagicMock

        sys.modules[imp] = MagicMock()
        _mocked_imports.append(imp)

from rdkit import RDLogger

RDLogger.DisableLog("rdApp.*")

import logging

logging.getLogger().setLevel(logging.ERROR)

import os
from os import path

import json

import pandas as pd

if not path.exists(path.join(path.expanduser("~"), f".oce/")):
    os.mkdir(path.join(path.expanduser("~"), f".oce/"))
    
CONFIG_PATH = os.path.join(os.path.expanduser("~"), ".oce/CONFIG.json")
if os.path.exists(CONFIG_PATH):
    with open(CONFIG_PATH) as f:
        CONFIG_ = json.load(f)
else:
    CONFIG_ = {}
    
defaults = {
    "MAP_LOCATION": "cpu",
    "USE_CUDA": False,
    "CDD_TOKEN": None,
    "VAULT_ID": None,
    "NUM_WORKERS": 0,
    "CACHE_PATH": os.path.join(os.path.expanduser("~"), ".oce/cache/"),
    "CACHE": True
}

for k, v in defaults.items():
    if k not in CONFIG_:
        CONFIG_[k] = v
        
with open(CONFIG_PATH, "w+") as f:
    json.dump(CONFIG_, f)

CONFIG = CONFIG_.copy()

if not path.exists(CONFIG["CACHE_PATH"]):
    os.mkdir(CONFIG["CACHE_PATH"])

if not path.exists(path.join(CONFIG["CACHE_PATH"], "vecrep/")):
    os.mkdir(path.join(CONFIG["CACHE_PATH"], "vecrep/")) # The passed CACHE_PATH must already exist

def update_config():
    """Update the configuration file.

    This function is called when a new parameter is added to the configuration file.
    """

    global CONFIG
    CONFIG = CONFIG_.copy()

    with contextlib.suppress(ImportError):
        import torch

        CONFIG["DEVICE"] = torch.device(CONFIG["MAP_LOCATION"])
        CONFIG["USE_CUDA"] = "cuda" in CONFIG["MAP_LOCATION"]

    with open(CONFIG_PATH, "w+") as f:
        json.dump(CONFIG_, f)


def set_config_param(param, value):
    """Set a configuration parameter.

    Parameters:
        param: the parameter to set.
        value: the value to set the parameter to.
    """
    CONFIG_[param] = value
    update_config()


def remove_config_param(param):
    """Remove a configuration parameter.

    Parameters:
        param: the parameter to remove.
    """
    CONFIG_.pop(param)
    update_config()


update_config()


def ExampleDataFrame():
    return pd.read_csv(
        "https://storage.googleapis.com/oloren-public-data/sample-csvs/sample_data3.csv"
    )


from .base_class import *
from .basics import *
from .beta import *
from .benchmarks import *
from .dataset import *
from .ensemble import *
from .external import *
from .gnn import *
from .hyperparameters import *
from .internal import *
from .interpret import *
from .manager import *
from .reduction import *
from .representations import *
from .splitters import *
from .uncertainty import *
from .visualizations import *


def ExampleDataset():
    dataset = (
        BaseDataset(
            data=ExampleDataFrame().to_csv(),
            structure_col="Smiles",
            property_col="pChEMBL Value",
        )
        + RandomSplit()
    )
    return dataset


def BACEDataset():
    df = pd.read_csv(download_public_file("MoleculeNet/load_bace_regression.csv"))
    df["split"] = df["split"].replace(
        {"Train": "train", "Valid": "valid", "Test": "test"}
    )
    return oce.BaseDataset(
        data=df.to_csv(), structure_col="smiles", property_col="pIC50"
    )


def test_oce():
    """Convenience function to test all functions of the oce package."""

    df = ExampleDataFrame()

    model = BaseBoosting(
        [
            RandomForestModel(
                DescriptastorusDescriptor("rdkit2dnormalized"), n_estimators=1
            ),
            BaseTorchGeometricModel(
                TLFromCheckpoint("default"), batch_size=8, epochs=1, preinitialized=True
            ),
            RandomForestModel(OlorenCheckpoint("default"), n_estimators=1),
        ]
    )

    model.fit(df["Smiles"], df["pChEMBL Value"])
    _ = model.predict(["CC(=O)OC1=CC=CC=C1C(=O)O"])
    save(model, "model.oce")
    _ = load("model.oce")
    os.remove("model.oce")

global online_session
online_session = None

def MISSING_DEPENDENCIES():
    if len(_mocked_imports) == 0:
        print("No missing dependencies.")
        return _mocked_imports
    else:
        print(textwrap.fill(f"Missing dependencies: {', '.join([x for x in _mocked_imports if '.' not in x])}.", 120))
        print("\nTo install OCE with all dependencies, run:")
        print("     bash <(curl -s https://raw.githubusercontent.com/Oloren-AI/olorenchemengine/master/install.sh)")

if len(_mocked_imports) > 0:
    import textwrap
    message = f"Some missing imports detected: {(','.join(_mocked_imports)[:30])}..." + \
    f"Automatically triggering online mode, where all OCE code is compiled locally and sent to AWS for execution." + \
    f"If you would like to use OCE online mode with private data or hosted on your own infrastructure, email." + \
    f" contact@oloren.ai for enterprise offerings." + \
    f"To exit online mode, run oce.online_session.__exit__()"

    message = f"""
Some missing imports detected: {(','.join(_mocked_imports)[:30])}...
Automatically triggering online mode, where all OCE code is compiled locally
and sent to AWS for execution. If you would like to use OCE online mode with private
data or hosted on your own infrastructure, email contact@oloren.ai for enterprise
offerings. To exit online mode, run olorenchemengine.online_session.__exit__()
""".replace("\n", " ")

    message = f"""
        To complete installation of Oloren ChemEngine, either:
            (1) Run oce.online() to use the demonstration package, Oloren ChemEngine Online,
            (2) Install the missing dependencies, instructions can be found oce.MISSING_DEPENDENCIES(), or
            (3) Email contact@oloren.ai with subject "Oloren ChemEngine Enterprise", for a secure privately hosted
                Server version.

        Oloren ChemEngine online is the public, dependency-free version of OCE, which compiles OCE code locally for fast,
        parallelized, remote execution on Oloren's cloud solution. Oloren ChemEngine online SHOULD NOT BE USED FOR
        CONFIDENTAIL DATA, and is only intended for demonstration purposes. The securely privately hosted Server version is
        called Oloren ChemEngine Enterprise.\n"""

    # print(textwrap.fill(textwrap.dedent(message).replace("\n", " ").replace("    (", "\n     ("), 80))
    print(textwrap.dedent(message))

def online(session_url="https://aws.chemengine.org"):
    global online_session
    online_session = Remote(session_url)
    online_session.__enter__()



import numpy as np
from rdkit import Chem
from rdkit.Chem.rdchem import BondType as BT
from torch_geometric.data import DataLoader
from tqdm import tqdm

import olorenchemengine as oce
from olorenchemengine.base_class import BaseModel, log_arguments
from olorenchemengine.internal import download_public_file
from olorenchemengine.representations import (
    AtomFeaturizer,
    BaseVecRepresentation,
    BondFeaturizer,
    SMILESRepresentation,
    TorchGeometricGraph,
)

ATOM_LIST = list(range(1, 119))
CHIRALITY_LIST = [
    Chem.rdchem.ChiralType.CHI_UNSPECIFIED,
    Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CW,
    Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CCW,
    Chem.rdchem.ChiralType.CHI_OTHER,
]
BOND_LIST = [BT.SINGLE, BT.DOUBLE, BT.TRIPLE, BT.AROMATIC]
BONDDIR_LIST = [
    Chem.rdchem.BondDir.NONE,
    Chem.rdchem.BondDir.ENDUPRIGHT,
    Chem.rdchem.BondDir.ENDDOWNRIGHT,
]


class MolCLR_AF(AtomFeaturizer):
    @property
    def length(self):
        return 2

    def convert(self, atom: Chem.Atom):
        atom_feature = [
            ATOM_LIST.index(atom.GetAtomicNum()),
            CHIRALITY_LIST.index(atom.GetChiralTag()),
        ]
        x = np.array(atom_feature)
        return x


class MolCLR_BF(BondFeaturizer):
    @property
    def length(self):
        return 2

    def convert(self, bond: Chem.Bond):
        edge_feature = [
            BOND_LIST.index(bond.GetBondType()),
            BONDDIR_LIST.index(bond.GetBondDir()),
        ]
        x = np.array(edge_feature)
        return x


class MolCLR_PYG(TorchGeometricGraph):
    @log_arguments
    def __init__(self):
        super().__init__(MolCLR_AF(), MolCLR_BF(), log=False)


class MolCLR(BaseModel):

    model_config = {
        "num_layer": 5,  # number of graph conv layers
        "emb_dim": 300,  # embedding dimension in graph conv layers
        "feat_dim": 512,  # output feature dimention
        "drop_ratio": 0.3,  # dropout ratio
        "pool": "mean",
    }

    @log_arguments
    def __init__(
        self,
        model_type="ginet",
        epochs=100,
        batch_size=32,
        init_lr=0.0005,
        init_base_lr=0.0001,
        weight_decay=1e-6,
        **kwargs,
    ):
        self.model_type = model_type

        self.config = {
            "epochs": epochs,
            "batch_size": batch_size,
            "init_lr": init_lr,
            "init_base_lr": init_base_lr,
            "weight_decay": weight_decay,
        }

        self.representation = MolCLR_PYG()

        super().__init__(**kwargs)

    def preprocess(self, X, y, **kwargs):
        if y is None:
            y = [None] * len(X)
        return self.representation.convert(X, ys=y)

    def _fit(self, X, y=None, **kwargs):

        import torch
        import torch.nn.functional as F
        from sklearn.metrics import (
            mean_absolute_error,
            mean_squared_error,
            roc_auc_score,
        )
        from torch import nn
        from torch.optim.lr_scheduler import CosineAnnealingLR
        from torch.utils.tensorboard import SummaryWriter

        if self.setting == "regression":
            self.criterion = nn.MSELoss()
        elif self.setting == "classification":
            self.criterion = nn.BCEWithLogitsLoss(reduction="none")

        if self.model_type == "ginet":
            from .model import GINet

            self.model = GINet(self.setting, **self.model_config)
            save_path = download_public_file("MolCLR/pretrained_gin.pth")
        elif self.model_type == "gcn":
            from .model import GCN

            self.model = GCN(self.setting, **self.model_config)
            save_path = download_public_file("MolCLR/pretrained_gcn.pth")

        self.model.load_my_state_dict(
            torch.load(save_path, map_location=oce.CONFIG["MAP_LOCATION"])
        )
        self.model.to(oce.CONFIG["DEVICE"])

        layer_list = []
        for name, param in self.model.named_parameters():
            if "pred_lin" in name:
                layer_list.append(name)

        params = list(
            map(
                lambda x: x[1],
                list(
                    filter(
                        lambda kv: kv[0] in layer_list, self.model.named_parameters()
                    )
                ),
            )
        )
        base_params = list(
            map(
                lambda x: x[1],
                list(
                    filter(
                        lambda kv: kv[0] not in layer_list,
                        self.model.named_parameters(),
                    )
                ),
            )
        )

        optimizer = torch.optim.Adam(
            [
                {"params": base_params, "lr": self.config["init_base_lr"]},
                {"params": params},
            ],
            self.config["init_lr"],
            weight_decay=self.config["weight_decay"],
        )

        train_loader = DataLoader(
            X,
            batch_size=self.config["batch_size"],
            shuffle=True,
            num_workers=oce.CONFIG["NUM_WORKERS"],
        )
        for epoch_counter in range(self.config["epochs"]):
            print(f"Epoch {epoch_counter+1}/{self.config['epochs']}")
            for bn, batch in enumerate(train_loader):

                batch = batch.to(oce.CONFIG["DEVICE"])
                _, pred = self.model(batch)
                y = batch.y.view(pred.shape).to(torch.float64)

                # Loss matrix
                loss_mat = self.criterion(pred.double(), y)

                optimizer.zero_grad()

                loss = torch.mean(loss_mat)
                loss.backward()

                optimizer.step()

    def _step(self, model, data, n_iter):
        # get the prediction
        __, pred = model(data)  # [N,C]
        loss = self.criterion(pred.flatten(), data.y)

        return loss

    def _predict(self, X, **kwargs):
        import torch

        loader = DataLoader(
            X,
            batch_size=self.config["batch_size"],
            shuffle=False,
            num_workers=oce.CONFIG["NUM_WORKERS"],
        )
        self.model.eval()
        y_pred = []

        for step, batch in enumerate(tqdm(loader, desc="Iteration")):
            batch = batch.to(oce.CONFIG["DEVICE"])

            with torch.no_grad():
                _, pred = self.model(batch)
            y_pred.append(pred.cpu().detach().numpy())

        return np.vstack(y_pred).flatten()


class MolCLRVecRep(BaseVecRepresentation):

    model_config = {
        "num_layer": 5,  # number of graph conv layers
        "emb_dim": 300,  # embedding dimension in graph conv layers
        "feat_dim": 512,  # output feature dimention
        "drop_ratio": 0.3,  # dropout ratio
        "pool": "mean",
    }

    @log_arguments
    def __init__(
        self,
        model_type="ginet",
        epochs=100,
        batch_size=32,
        init_lr=0.0005,
        init_base_lr=0.0001,
        weight_decay=1e-6,
        **kwargs,
    ):
        self.model_type = model_type

        self.config = {
            "epochs": epochs,
            "batch_size": batch_size,
            "init_lr": init_lr,
            "init_base_lr": init_base_lr,
            "weight_decay": weight_decay,
        }

        import torch
        import torch.nn.functional as F
        from sklearn.metrics import (
            mean_absolute_error,
            mean_squared_error,
            roc_auc_score,
        )
        from torch import nn
        from torch.optim.lr_scheduler import CosineAnnealingLR
        from torch.utils.tensorboard import SummaryWriter

        if self.model_type == "ginet":
            from .model import GINet

            self.model = GINet("regression", **self.model_config)
            save_path = download_public_file("MolCLR/pretrained_gin.pth")
        elif self.model_type == "gcn":
            from .model import GCN

            self.model = GCN("regression", **self.model_config)
            save_path = download_public_file("MolCLR/pretrained_gcn.pth")

        self.model.load_my_state_dict(
            torch.load(save_path, map_location=oce.CONFIG["MAP_LOCATION"])
        )
        self.model.to(oce.CONFIG["DEVICE"])

    def _convert(self, X, **kwargs):
        return self.convert(X)

    def convert(self, X, **kwargs):
        X = MolCLR_PYG().convert(X, ys=None)
        import torch

        loader = DataLoader(
            X,
            batch_size=self.config["batch_size"],
            shuffle=True,
            num_workers=oce.CONFIG["NUM_WORKERS"],
        )
        self.model.eval()
        y_pred = []

        for step, batch in enumerate(tqdm(loader, desc="Iteration")):
            batch = batch.to(oce.CONFIG["DEVICE"])

            with torch.no_grad():
                rep, _ = self.model(batch)
            y_pred.append(rep.cpu().detach().numpy())

        return np.concatenate(y_pred)



"""
These visualizations are intended for exploratory analysis.
"""

from olorenchemengine.base_class import *
from olorenchemengine.dataset import *
from olorenchemengine.internal import *
from olorenchemengine.interpret import *
from olorenchemengine.manager import *
from olorenchemengine.representations import *
from olorenchemengine.uncertainty import *
from olorenchemengine.visualizations import *


class VisualizeFeatures(BaseVisualization):
    """
    Visualize how the selected features are correlated with the dataset.

    Parameters:
        dataset (BaseDataset): the dataset to analyze
        features (list): the features to visualize. For the elements of the feature
            list, strings will be treated as a column of the dataset and BaseVecRepresentations
            are used to convert structures to features, the provided representations
            should have the names property defined for better interpretability.
            In addition to features thefeature_cols in the dataset are also visualized."""

    @log_arguments
    def __init__(self, dataset: BaseDataset, features: list = [], log=True, **kwargs):
        self.dataset = dataset
        self.features = dataset.feature_cols + features
        self.packages = ["plotly", "smilesdrawer"]
        self.data = {
            "SMILES": self.dataset.data[self.dataset.structure_col].tolist(),
            self.dataset.property_col: self.dataset.data[
                self.dataset.property_col
            ].tolist(),
        }
        self.d = {}
        self.feature_cols = []
        self.spearman_corr = []
        from scipy.stats import spearmanr

        for feature in self.features:
            if isinstance(feature, str):
                self.feature_cols.append(features)
                self.d[feature] = self.dataset.data[feature].tolist()
                self.spearman_corr.append(
                    spearmanr(
                        self.d[feature],
                        self.dataset.data[self.dataset.property_col].tolist(),
                    )
                )
            elif issubclass(type(feature), BaseRepresentation):
                rep_features = feature.convert(self.dataset.entire_dataset[0])
                rep_features = np.array(rep_features).transpose()
                try:
                    names = feature.names
                except:
                    names = [
                        f"{feature.__class__.__name__}-{i}"
                        for i in range(rep_features.shape[0])
                    ]
                for i, name in enumerate(names):
                    self.feature_cols.append(name)
                    self.d[name] = rep_features[i].tolist()
                    self.spearman_corr.append(
                        spearmanr(
                            self.d[name], self.dataset.data[self.dataset.property_col]
                        )
                    )

    def get_data(self):
        d = {
            "datacols": self.d.copy(),
            "SMILES": self.data["SMILES"],
            "PROPERTY": self.dataset.property_col,
            "PROPERTY_VALUES": self.data[self.dataset.property_col],
            "FEATURE_COLS": self.feature_cols,
            "SPEARMAN_COEF": [np.around(x[0], decimals=2) for x in self.spearman_corr],
            "SPEARMAN_PVAL": [np.around(x[1], decimals=2) for x in self.spearman_corr],
            "hoverSize": 0.8,
        }
        return d


import os
from typing import Any, List, Union

import numpy as np
from rdkit import Chem

import olorenchemengine as oce
from olorenchemengine.base_class import log_arguments
from olorenchemengine.internal import download_public_file
from olorenchemengine.representations import BaseVecRepresentation

from .operations import DfVec, MolSentence, mol2alt_sentence, sentences2vec


class Mol2Vec(BaseVecRepresentation):
    @log_arguments
    def __init__(self):
        print(
            "Requires gensim installation https://github.com/RaRe-Technologies/gensim"
        )

        model_path = download_public_file("mol2vec/model_300dim.pkl")

        oce.import_or_install("gensim")

        from gensim.models import word2vec

        self.model = word2vec.Word2Vec.load(model_path)
        super().__init__(log=False)

    def _convert(self, s: str) -> np.ndarray:
        mol = Chem.MolFromSmiles(s)
        sentence = MolSentence(mol2alt_sentence(mol, 1))
        X = DfVec(sentences2vec([sentence], self.model, unseen="UNK")[0]).vec
        return np.array(X)

    def _convert_list(
        self, smiles_list: List[str], ys: List[Union[int, float, np.number]] = None
    ) -> List[np.ndarray]:
        mols = [Chem.MolFromSmiles(s) for s in smiles_list]
        sentences = [MolSentence(mol2alt_sentence(mol, 1)) for mol in mols]
        X = [
            DfVec(sentence).vec
            for sentence in sentences2vec(sentences, self.model, unseen="UNK")
        ]
        return np.array(X)



import olorenchemengine as oce
from olorenchemengine.internal import *
from olorenchemengine.gnn import *

import torch.nn.functional as F

from torch_geometric.nn import SuperGATConv

class SuperGATModel_beta(BaseLightningModule):

    """ SuperGAT is a network

    Parameters:
        dropout (float, optional): The dropout rate to use for the model. Default is 0.1.
        lr (float, optional): The learning rate to use for training. Default is 1e-4.
        optim (str, optional): The optimizer to use for training. Default is "adam".
    """

    @log_arguments
    def __init__(
        self,
        hidden_channels: int = 8,
        heads: int = 8,
        pooling_function: str = "mean",
        negative_slope: float = 0.2,
        dropout: float = 0.0,
        add_self_loops: bool = True,
        bias: bool = True,
        attention_type: str = "MX",
        neg_sample_ratio: float = 0.5,
        edge_sample_ratio: float = 1.0,
        is_undirected: bool = True,
        lr: float = 1e-4,
        optim: str = "adam",
    ):
        self.lr = lr
        super().__init__(optim=optim)
        
        self.hidden_channels = hidden_channels
        self.heads = heads
        self.negative_slope = negative_slope
        self.dropout = dropout
        self.add_self_loops = add_self_loops
        self.bias = bias
        self.attention_type = attention_type
        self.neg_sample_ratio = neg_sample_ratio
        self.edge_sample_ratio = edge_sample_ratio
        self.is_undirected = is_undirected
        self.lr = lr
        self.optim = optim
        
        from torch_geometric.nn import (
            global_add_pool,
            global_max_pool,
            global_mean_pool,
        )

        if pooling_function.lower() == "add":
            self.pooling_function = global_add_pool
        elif pooling_function.lower() == "mean":
            self.pooling_function = global_mean_pool
        elif pooling_function.lower() == "max":
            self.pooling_function = global_max_pool

    def create(self, dimensions):
        self.conv1 = SuperGATConv(dimensions[0], self.hidden_channels, heads=self.heads,
                                  dropout=self.dropout, attention_type=self.attention_type,
                                  edge_sample_ratio=self.edge_sample_ratio, is_undirected=self.is_undirected,
                                  negative_slope=self.negative_slope, add_self_loops=self.add_self_loops,
                                  bias=self.bias, neg_sample_ratio=self.neg_sample_ratio)
        self.conv2 = SuperGATConv(self.hidden_channels*self.heads, self.hidden_channels*self.heads,
                                  heads=self.heads,
                                  concat=False, dropout=self.dropout, attention_type=self.attention_type,
                                  edge_sample_ratio=self.edge_sample_ratio, is_undirected=self.is_undirected,
                                  negative_slope=self.negative_slope, add_self_loops=self.add_self_loops,
                                  bias=self.bias, neg_sample_ratio=self.neg_sample_ratio)
        self.lin1 = nn.Linear(self.hidden_channels*self.heads, 1)

    def forward(self, batch):
        if batch.x.is_cuda:
            dtype = torch.cuda.FloatTensor
        else:
            dtype = torch.FloatTensor
        x = batch.x.type(dtype)
        
        edge_index = batch.edge_index

        del dtype
        
        x = F.dropout(x, p=self.dropout, training=self.training)
        x = F.elu(self.conv1(x, edge_index))
        x = F.dropout(x, p=self.dropout, training=self.training)
        x = self.conv2(x, edge_index)
        x = self.pooling_function(x, batch.batch, size=batch.num_graphs)
        return self.lin1(x)
    
    ##########################
    # train, val, test, steps
    ##########################
    def training_step(self, batch, batch_idx):
        import torch.nn.functional as F
        from torch_geometric.nn import global_mean_pool

        if batch.x.is_cuda:
            dtype = torch.cuda.FloatTensor
        else:
            dtype = torch.FloatTensor
        x = batch.x.type(dtype)
        
        edge_index = batch.edge_index
        #edge_attr = batch.edge_attr.type(dtype)

        #del dtype
        
        x = F.dropout(x, p=0.6, training=self.training)
        x = F.elu(self.conv1(x, edge_index))
        att_loss = self.conv1.get_attention_loss()
        x = F.dropout(x, p=0.6, training=self.training)
        
        x = self.conv2(x, edge_index)
        att_loss += self.conv2.get_attention_loss()
        x = self.pooling_function(x, batch.batch, size=batch.num_graphs)
        y_pred = self.lin1(x)
        
        loss = self.loss(y_pred, batch.y)
        loss += 4.0 * att_loss
        return loss
        
    def validation_step(self, batch, batch_idx):
        return self.training_step(batch, batch_idx)
    
    def test_step(self, batch, batch_idx):
        return self.training_step(batch, batch_idx)


import copy
import io
from abc import abstractmethod

import numpy as np
import PIL
import selfies as sf
from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.DataStructs.cDataStructs import BulkTanimotoSimilarity, TanimotoSimilarity
from tqdm import tqdm

from .base_class import *
from .dataset import *
from .external.stoned import *
from .representations import *


class PerturbationEngine(BaseClass):

    """PerturbationEngine is the base class for techniques which mutate or perturb a compound
    into a similar one with a small difference.

    Methods:
        get_compound_at_idx: returns a compound with a modification at a given
            atom index
        get_compound: returns a compound with a randomly chosen modification
        get_compound_list: returns a list of compounds with modifications, the list
            is meant to be comprehensive of the result of the application of an
            entire class of modifications."""

    @abstractmethod
    def get_compound_at_idx(self, mol: Chem.Mol, idx: int) -> str:
        pass

    @abstractmethod
    def get_compound(self, smiles, n=1, **kwargs) -> str:
        pass

    @abstractmethod
    def get_compound_list(self, smiles, **kwargs) -> list:
        pass


class SwapMutations(PerturbationEngine):
    """SwapMutations replaces substructures with radius r with another substructure
    with radius < r. The substructure is chosen such that it has the same outgoing
    bonds, and this set of substructures is identified through a comprehensive
    ennumeration of a large set of lead-like compounds.

    Methods:
        get_compound: returns a compound with a randomly chosen modification
        get_compound_list: returns a list of compounds with modifications, the list
            is meant to be comprehensive of the result of the application of an
            entire class of modifications"""

    @log_arguments
    def __init__(self, radius=0, log=True):
        self.radius = radius

        if not radius in (0, 1, 2):
            raise ValueError("radius must be 0, 1, or 2")

        self.trans = dict()
        for radius in range(radius + 1):
            transformation_path = download_public_file(
                f"swap-mutations/trans_{radius}.json"
            )
            with open(transformation_path, "r") as f:
                self.trans.update(json.load(f))

    def get_substitution(self, m, idx, r=1):

        # finding atoms within radius r of atom idx
        substructure_atoms = set({idx})
        for i in range(r):
            bonds = [
                b for a_ in substructure_atoms for b in m.GetAtomWithIdx(a_).GetBonds()
            ]
            substructure_atoms_ = {b.GetBeginAtomIdx() for b in bonds}.union(
                {b.GetEndAtomIdx() for b in bonds}
            )
            substructure_atoms = substructure_atoms.union(substructure_atoms_)

        # finding atoms connecting to substructure
        surface_atoms = {
            b.GetEndAtomIdx()
            for a_ in substructure_atoms
            for b in m.GetAtomWithIdx(a_).GetBonds()
            if not b.GetEndAtomIdx() in substructure_atoms
        }
        surface_atoms.update(
            {
                b.GetBeginAtomIdx()
                for a_ in substructure_atoms
                for b in m.GetAtomWithIdx(a_).GetBonds()
                if not b.GetBeginAtomIdx() in substructure_atoms
            }
        )
        surface_atoms = list(surface_atoms)
        surface_l = list()
        for surface_atom in surface_atoms:
            bonds = [
                b.GetIdx()
                for b in m.GetAtomWithIdx(surface_atom).GetBonds()
                if b.GetEndAtomIdx() in substructure_atoms
                or b.GetBeginAtomIdx() in substructure_atoms
            ]
            surface_l.append(
                (
                    m.GetAtomWithIdx(surface_atom).GetAtomicNum(),
                    [m.GetBondWithIdx(b).GetBondTypeAsDouble() for b in bonds],
                )
            )

        tmp = [
            (x1, x2, x3)
            for x1, x2, x3 in zip(
                [sum(surface_b) for surface_a, surface_b in surface_l],
                surface_atoms,
                surface_l,
            )
        ]
        tmp.sort(key=lambda x: x[0])
        surface_l = [x for _, _, x in tmp]
        surface_atoms = [x for _, x, _ in tmp]

        return surface_l, surface_atoms, substructure_atoms

    def get_entry(self, m, idx, r=1):
        m = copy.deepcopy(m)

        # finding atoms within radius r of atom idx
        substructure_atoms = set({idx})
        for i in range(r):
            bonds = [
                b for a_ in substructure_atoms for b in m.GetAtomWithIdx(a_).GetBonds()
            ]
            substructure_atoms_ = {b.GetBeginAtomIdx() for b in bonds}.union(
                {b.GetEndAtomIdx() for b in bonds}
            )
            substructure_atoms = substructure_atoms.union(substructure_atoms_)

        # finding atoms connecting to substructure
        surface_atoms = {
            b.GetEndAtomIdx()
            for a_ in substructure_atoms
            for b in m.GetAtomWithIdx(a_).GetBonds()
            if not b.GetEndAtomIdx() in substructure_atoms
        }
        surface_atoms.update(
            {
                b.GetBeginAtomIdx()
                for a_ in substructure_atoms
                for b in m.GetAtomWithIdx(a_).GetBonds()
                if not b.GetBeginAtomIdx() in substructure_atoms
            }
        )
        surface_atoms = list(surface_atoms)
        surface_l = list()
        for surface_atom in surface_atoms:
            bonds = [
                b.GetIdx()
                for b in m.GetAtomWithIdx(surface_atom).GetBonds()
                if b.GetEndAtomIdx() in substructure_atoms
                or b.GetBeginAtomIdx() in substructure_atoms
            ]
            surface_l.append(
                (
                    m.GetAtomWithIdx(surface_atom).GetAtomicNum(),
                    [m.GetBondWithIdx(b).GetBondTypeAsDouble() for b in bonds],
                )
            )

        tmp = [
            (x1, x2, x3)
            for x1, x2, x3 in zip(
                [sum(surface_b) for surface_a, surface_b in surface_l],
                surface_atoms,
                surface_l,
            )
        ]
        tmp.sort(key=lambda x: x[0])
        surface_l = [x for _, _, x in tmp]
        surface_atoms = [x for _, x, _ in tmp]

        for i, j in enumerate(surface_atoms):
            m.GetAtomWithIdx(j).SetAtomMapNum(i + 1)

        substruct = Chem.RWMol(m)

        removal = list()
        for i in range(substruct.GetNumAtoms()):
            if i not in substructure_atoms and i not in surface_atoms:
                removal.append(substruct.GetAtomWithIdx(i))

        for atom in removal:
            substruct.RemoveAtom(atom.GetIdx())
        return surface_l, substruct.GetMol()

    def stitch(self, m):
        m = Chem.RWMol(m)
        mappings = {}
        for a in m.GetAtoms():
            if a.GetAtomMapNum() != 0:
                if not a.GetAtomMapNum() in mappings:
                    mappings[a.GetAtomMapNum()] = list()
                mappings[a.GetAtomMapNum()].append(a)
        removal = []
        for k, v in mappings.items():
            i = m.AddAtom(Chem.Atom(v[0].GetAtomicNum()))
            a = m.GetAtomWithIdx(i)

            added = []
            for v_ in v:
                for b in v_.GetBonds():
                    if v_.GetIdx() == b.GetEndAtomIdx():
                        i = b.GetBeginAtomIdx()
                    else:
                        i = b.GetEndAtomIdx()
                    if not i in added:
                        m.AddBond(a.GetIdx(), i, b.GetBondType())
                        added.append(i)

                removal.append(v_)

        for remove in removal:
            m.RemoveAtom(remove.GetIdx())
        return m

    def get_compound_at_idx(self, mol, idx, **kwargs):
        ref_m = mol

        l, a, sub_a = self.get_substitution(ref_m, idx, r=self.radius)
        # if the current substructure to replace has no replacements in the database
        if str(l) not in self.trans:
            return None

        for x in sub_a:
            ref_m.GetAtomWithIdx(x).SetAtomMapNum(10000)

        for k, j in enumerate(a):
            ref_m.GetAtomWithIdx(j).SetAtomMapNum(k + 1)

        sub = np.random.choice(self.trans[str(l)])
        m = Chem.CombineMols(ref_m, Chem.MolFromSmiles(sub, sanitize=False))

        m = Chem.RWMol(m)

        removal = []
        for i in range(len(m.GetAtoms())):
            if m.GetAtomWithIdx(i).GetAtomMapNum() == 10000:
                removal.append(m.GetAtomWithIdx(i))

        for remove in removal:
            m.RemoveAtom(remove.GetIdx())
        if m is None:
            errors += 1
        try:
            s1 = Chem.MolToSmiles(m)
            m = self.stitch(m)
            if not m is None:
                return Chem.MolToSmiles(m)
        except Exception as e:
            return None

    def get_compound(self, smiles, **kwargs):
        if isinstance(smiles, Chem.Mol):
            mol = smiles
            smiles = Chem.MolToSmiles(mol)
        else:
            mol = Chem.MolFromSmiles(smiles)
        idx = np.random.choice(mol.GetNumAtoms(), replace=False)
        return self.get_compound_at_idx(mol, idx, **kwargs)

    def get_compound_list(self, smiles, idx: int = None, **kwargs) -> list:
        outs = []
        ref_m = Chem.MolFromSmiles(smiles)
        if idx is None:
            space = range(len(ref_m.GetAtoms()))
        else:
            space = [idx]
        for i in tqdm(space):
            ref_m = Chem.MolFromSmiles(smiles)
            l, a, sub_a = self.get_substitution(ref_m, i, r=self.radius)

            for x in sub_a:
                ref_m.GetAtomWithIdx(x).SetAtomMapNum(10000)

            for k, j in enumerate(a):
                ref_m.GetAtomWithIdx(j).SetAtomMapNum(k + 1)

            surface_atoms = [ref_m.GetAtomWithIdx(j) for j in a]
            if str(l) not in self.trans.keys():
                print("Continuing")
                continue
            for sub in self.trans[str(l)]:
                m = Chem.CombineMols(ref_m, Chem.MolFromSmiles(sub, sanitize=False))

                m = Chem.RWMol(m)

                removal = []
                for i in range(len(m.GetAtoms())):
                    if m.GetAtomWithIdx(i).GetAtomMapNum() == 10000:
                        removal.append(m.GetAtomWithIdx(i))

                for remove in removal:
                    m.RemoveAtom(remove.GetIdx())
                try:
                    m = self.stitch(m)
                    s = Chem.MolToSmiles(m)
                    m = Chem.MolFromSmiles(s)
                    if not m is None:
                        outs.append(s)
                except Exception as e:
                    pass
        return list(set(outs))

    def _save(self) -> dict:
        return super()._save()

    def _load(self, d: dict):
        return super()._load(d)


class STONEDMutations(PerturbationEngine):
    """Implements STONED-SELFIES algorithm for generating modified compounds.

    `STONED-SELFIES GitHub repository <https://github.com/aspuru-guzik-group/stoned-selfies>`_
    `Beyond Generative Models: Superfast Traversal, Optimization, Novelty, Exploration and Discovery
    (STONED) Algorithm for Molecules using SELFIES <https://doi.org/10.26434/chemrxiv.13383266.v2>`_

    Methods:
        get_compound_at_idx: returns a compound with a randomly chosen modification at `idx`
        get_compound: returns a compound with a randomly chosen modification
        get_compound_list: returns a list of `num_samples` compounds with randomly chosen modifications
    """

    @log_arguments
    def __init__(self, mutations: int = 1, log=True):
        self.mutations = mutations

    def get_compound_at_idx(self, mol: Chem.Mol, idx: int, **kwargs) -> str:
        smiles = Chem.MolToSmiles(mol)
        selfie = sf.encoder(smiles)
        selfie_chars = get_selfie_chars(selfie)
        max_molecules_len = len(selfie_chars) + self.mutations
        selfie_mutated, _ = mutate_selfie(selfie, max_molecules_len, index=idx)
        smiles_back = sf.decoder(selfie_mutated)
        return smiles_back

    def get_compound(self, smiles: str, **kwargs) -> str:
        ref_m = Chem.MolFromSmiles(smiles)
        randomized_smiles_ordering = randomize_smiles(ref_m)
        selfie = sf.encoder(randomized_smiles_ordering)
        selfie_mutated = get_mutated_SELFIES([selfie], num_mutations=self.mutations)[0]
        smiles_back = sf.decoder(selfie_mutated)
        return smiles_back

    def get_compound_list(self, smiles: str, num_samples: int = 1000, **kwargs) -> list:
        ref_m = Chem.MolFromSmiles(smiles)
        randomized_smile_orderings = [
            randomize_smiles(ref_m) for _ in range(num_samples)
        ]
        selfies_ls = [sf.encoder(x) for x in randomized_smile_orderings]
        selfies_mut = get_mutated_SELFIES(
            selfies_ls.copy(), num_mutations=self.mutations
        )
        smiles_back = [sf.decoder(x) for x in selfies_mut]
        return list(set(smiles_back))

    def _save(self) -> dict:
        return super()._save()

    def _load(self, d: dict):
        return super()._load(d)


class CounterfactualEngine(BaseClass):
    """Generates counterfactual compounds based on:

    `exmol GitHub repository <https://github.com/ur-whitelab/exmol>`_
    `Model agnostic generation of counterfactual explanations for molecules <http://dx.doi.org/10.1039/D1SC05259D>`_
    """

    @log_arguments
    def __init__(
        self, model: BaseModel, perturbation_engine: PerturbationEngine = "default"
    ):
        self.model = model
        if perturbation_engine == "default":
            self.perturbation_engine = SwapMutations(radius=1)
        else:
            self.perturbation_engine = perturbation_engine
        self.samples = []
        self.cfs = []

    def generate_samples(self, smiles: str) -> None:
        """Generates candidate counterfactuals and stores them in `self.samples` as a list of dictionaries.

        Args:
            smiles: SMILES string of the target prediction
        """
        perturbed_smiles = [smiles] + self.perturbation_engine.get_compound_list(smiles)
        filtered_smiles = []
        seen = set()
        for smi in tqdm(perturbed_smiles):
            mol, smi_canon, conversion_successful = sanitize_smiles(smi)
            if conversion_successful and smi_canon not in seen:
                filtered_smiles.append(smi_canon)
                seen.add(smi_canon)

        ref_m = Chem.MolFromSmiles(smiles)
        filtered_mols = [Chem.MolFromSmiles(s) for s in filtered_smiles]

        ref_fp = AllChem.GetMorganFingerprint(ref_m, 2)
        fps = [AllChem.GetMorganFingerprint(m, 2) for m in filtered_mols]
        scores = BulkTanimotoSimilarity(ref_fp, fps)

        values = self.model.predict(filtered_smiles)

        dmat = 1 - np.array([BulkTanimotoSimilarity(fp, fps) for fp in tqdm(fps)])

        from sklearn.cluster import DBSCAN
        from sklearn.decomposition import PCA

        coords = PCA(n_components=2).fit_transform(dmat)
        clustering = DBSCAN(eps=0.15, min_samples=5).fit(coords)

        samples = {
            "SMILES": filtered_smiles,
            "Similarity": scores,
            "Value": values,
            "Coordinate": list(coords),
            "Cluster": clustering.labels_,
            "Original": [True] + [False for _ in filtered_smiles[1:]],
        }
        self.samples = pd.DataFrame(samples).to_dict("records")

    def get_samples(self) -> pd.DataFrame:
        """Returns candidate counterfactuals as a pandas dataframe.

        Returns:
            pandas dataframe of candidate counterfactuals
        """
        return pd.DataFrame(self.samples)

    def generate_cfs(
        self, delta: Union[int, float, Tuple] = (-1, 1), n: int = 4
    ) -> None:
        """Generates counterfactuals and stores them in `self.cfs` as a list of dictionaries.

        Args:
            delta: margin defining counterfactuals for regression models
            n: number of counterfactuals
        """
        base = self.samples[0]["Value"]

        if self.model.setting == "classification":
            cfs = self._select_cfs(lambda s: s["Value"] != base, n)
            self.cfs = self.samples[:1] + cfs
        else:
            if type(delta) in (float, int):
                delta = (-delta, delta)
            lcfs = self._select_cfs(lambda s: s["Value"] - delta[0] < base, n // 2)
            hcfs = self._select_cfs(lambda s: s["Value"] - delta[1] > base, n // 2)
            self.cfs = self.samples[:1] + lcfs + hcfs

    def get_cfs(self) -> pd.DataFrame:
        """Returns counterfactuals as a pandas dataframe.

        Returns:
            pandas dataframe of counterfactuals
        """
        return pd.DataFrame(self.cfs)

    def _select_cfs(self, fn: Callable, n: int) -> list:
        """Selects counterfactuals from `self.samples` using a validity function.

        Args:
            fn: boolean function for identifying counterfactuals
            n: maximum number of counterfactuals
        Returns:
            cfs: list of dictionary of counterfactuals
        """
        cfs = []

        def cluster_score(s, i):
            """Score for clustering for a given model.

            Parameters:
                s: sample
                i: index of sample

            Returns:
                score: score for clustering
            """
            return (s["Cluster"] == i) * fn(s) * s["Similarity"]

        clusters = {s["Cluster"] for s in self.samples[1:]}
        for i in clusters:
            candidate = max(self.samples[1:], key=lambda s: cluster_score(s, i))
            if cluster_score(candidate, i):
                cfs.append(candidate)

        cfs = sorted(cfs, key=lambda s: s["Similarity"], reverse=True)[:n]
        return cfs

    def _save(self) -> dict:
        return super()._save()

    def _load(self, d: dict):
        return super()._load(d)


def model_molecule_sensitivity(
    model: BaseModel,
    smiles: str,
    perturbation_engine: PerturbationEngine = "default",
    n: int = 30,
) -> Chem.Mol:

    """
    Calculates the sensitivity of a model to perturbations in on each of a molecule's atoms,
    outputting a rdkit molecule, with sensitivity as an atom property.

    Parameters:
        model: model to be used for sensitivity calculation
        smiles: SMILES string of the target prediction
        perturbation_engine: perturbation engine to be used for sensitivity calculation
        n: number of perturbations to be used for sensitivity calculation

    Returns:
        rdkit molecule with sensitivity as an atom property
    """

    if perturbation_engine == "default":
        perturbation_engine = SwapMutations(radius=1)

    mol = Chem.MolFromSmiles(smiles)

    idxs = []

    for a in tqdm(mol.GetAtoms()):
        idx = a.GetIdx()
        smiles = [Chem.MolToSmiles(mol)]

        count = 1
        while count < n:
            perturbed_smiles = perturbation_engine.get_compound_at_idx(smiles[0], idx)
            try:
                mol2 = Chem.MolFromSmiles(perturbed_smiles)
                if not mol2 is None:
                    smiles.append(perturbed_smiles)
                    idxs.append(idx)
                    count += 1
            except:
                pass
        predictions = model.predict(smiles)
        v = np.std(predictions)
        a.SetProp("sensitivity", str(v))
        a.SetProp("atomNote", str(np.format_float_scientific(v, precision=2)))
    return mol



"""
for basic analysis of compounds/groups of compounds
"""

from olorenchemengine.base_class import *
from olorenchemengine.dataset import *
from olorenchemengine.internal import *
from olorenchemengine.interpret import *
from olorenchemengine.manager import *
from olorenchemengine.representations import *
from olorenchemengine.uncertainty import *
from olorenchemengine.visualizations import *

from typing import *

class VisualizeMCS(VisualizeCompounds):
    """
    Visualize the maximum common substructure between a list of compounds.
    
    Parameters:
        smiles (List[str]): the smiles of the compounds to visualize
        timeout (int): the timeout for the MCS calculation in seconds
        invert_colors (bool): invert the colors of the MCS such that the MCS is 
            red and the differences are green. Default False.
        completeRingsOnly (bool): only consider complete rings in the MCS, parameter
            for rdkit.Chem.rdFMCS.FindMCS. Default True.
    """
    
    @log_arguments
    def __init__(self, smiles: List[str], *args, timeout = 30, kekulize = False, 
            completeRingsOnly=True, invert_colors=False, log=True, **kwargs):
        mols = [Chem.MolFromSmiles(s) for s in smiles]
        mols_ = []
        for s, m in zip(smiles, mols):
            if m is None:
                print(f"Could not parse {s}, skipping")
            mols_.append(m)
        mols = mols_
        
        from rdkit.Chem.rdFMCS import FindMCS
        result = FindMCS(mols, completeRingsOnly=completeRingsOnly, timeout = timeout)
        smarts = result.smartsString
        for mol in mols:
            matches = mol.GetSubstructMatches(Chem.MolFromSmarts(smarts))
            for a_idx in matches[0]:
                mol.GetAtomWithIdx(a_idx).SetAtomMapNum(2)
            for a_idx in range(mol.GetNumAtoms()):
                if a_idx not in matches[0]:
                    mol.GetAtomWithIdx(a_idx).SetAtomMapNum(1)
        
        smiles_highlighted = [Chem.MolToSmiles(mol) for mol in mols]
        
        if invert_colors:
            highlights = [[2, "#fb8fff"], [1, "#8fff9c"]]
        else:
            highlights = [[1, "#fb8fff"], [2, "#8fff9c"]]
        super().__init__(smiles_highlighted, *args, highlights = highlights, kekulize=kekulize, log=False, **kwargs)

    @property
    def JS_NAME(self):
        return "VisualizeCompounds"


import olorenchemengine as oce

from .base_class import *

class PCA(BaseSKLearnReduction):
    
    """Wrapper for sklearn PCA"""

    @log_arguments
    def __init__(self, *args, **kwargs):
        from sklearn.decomposition import PCA

        self.obj = PCA(*args, **kwargs)
    
class FactorAnalysis(BaseSKLearnReduction):
    
    """Wrapper for sklearn FactorAnalysis"""

    @log_arguments
    def __init__(self, *args, **kwargs):
        from sklearn.decomposition import FactorAnalysis

        self.obj = FactorAnalysis(*args, **kwargs)


from olorenchemengine.internal import *
from olorenchemengine.representations import *

class SimLookup_beta(BaseCompoundVecRepresentation):
    
    @log_arguments
    def __init__(self, rep: BaseCompoundVecRepresentation, *args, metric: str = "cosine", 
                 k: int = 3, return_reps: bool = False, log: bool = True, **kwargs):
        self.rep = rep
        self.metric = metric
        self.k = k
        self.return_reps = return_reps
        
        self.fitted = False
        super().__init__(*args, log=False, **kwargs)
    
    def _convert(self):
        assert False, "Directly accessing _convert is not implemented with SimLookup. Use convert instead."
        
    def convert(self, Xs, ys = None, fit = False, **kwargs):
        Xs = SMILESRepresentation().convert(Xs)
        
        assert fit or self.fitted, "SimLookup has not been fitted yet. Call convert with fit=True first."
        
        if fit:
            self.Xs = Xs
            self.ys = ys
            dists = self.rep.calculate_distance(Xs, Xs, metric = self.metric)
            dists = dists + np.identity(dists.shape[0])
            best_ix = np.argpartition(dists, self.k, axis=1)[:, :self.k]
            feats = []
            for i in range(len(Xs)):
                if self.return_reps:
                    feats.append(np.concatenate([dists[i][best_ix[i]], ys[best_ix[i]], self.rep.convert([self.Xs[i]])[0]]))
                else:
                    feats.append(np.concatenate([dists[i][best_ix[i]], ys[best_ix[i]]]))
            self.fitted = True
            return feats
        else:
            dists = self.rep.calculate_distance(Xs, self.Xs, metric = self.metric)
            best_ix = np.argpartition(dists, self.k, axis=1)[:, :self.k]
            feats = []
            for i in range(len(Xs)):
                if self.return_reps:
                    feats.append(np.concatenate([dists[i][best_ix[i]], self.ys[best_ix[i]], self.rep.convert([self.Xs[i]])[0]]))
                else:
                    feats.append(np.concatenate([dists[i][best_ix[i]], self.ys[best_ix[i]]]))
            return feats
    
    def _save(self):
        d = super()._save()
        d.update({"fitted", self.fitted,})
        if hasattr(self, "Xs"):
            d.update({"Xs": self.Xs, "ys": self.ys})
        return d
    
    def _load(self, d):
        super()._load(d)
        self.fitted = d["fitted"]
        if "Xs" in d:
            self.Xs = d["Xs"]
            self.ys = d["ys"]
            


"""piCalculax is a library for isoelectric point prediction for chemically modified peptides.

The original library has been adapted to be incorporated into Oloren ChemEngine

For more information visit the github repository: https://github.com/EBjerrum/pICalculax
"""

from numpy import abs as np_abs
from numpy import argmin, unique

# RuleTables
from rdkit import Chem
from rdkit.Chem import AllChem

# Format of ruletable tuple (Name, SMARTS, pKa (evt. as list), type (evt.as list) )
# Type is 0 = acidic (charge below pKa)
# Type 1 is basic (charge below pKa)

# (Grimsley 2009) with extensions
# Grimsley, G. R.; Scholtz, J. M. & Pace, C. N. (2009), 'A summary of the measured pK values of the ionizable groups in folded proteins.', Protein science : a publication of the Protein Society 18, 247--251.

# Default pKa tables defined at the end of the file

pkatable_grimsley = [
    (
        "Tetramethylrhodamine #3",
        Chem.MolFromSmiles(
            "CN(C)C1=C-C=C2-C(C4=C(C(=O)O)C=CC=C4)=C3-C=C-C(=[N+](C)C)C=C3OC2=C1"
        ),
        [-10, 1.19, 3.59, 10],
        [1, 1, 0, 1],
    ),
    (
        "6-hydroxy-3H-xanthen-3-one",
        Chem.MolFromSmarts("c2c3ccc(O)cc3oc3cc(=O)ccc2-3"),
        [2.94, 6.56],
        [1, 0],
    ),
    ("Anilin", Chem.MolFromSmarts("[$([NH2]c1ccccc1)]"), 4.87, 1),
    ("N-acetyl Hydrazine", Chem.MolFromSmarts("[$([NH2]NC(=O))]"), 2.81, 1),  # Reaxys
    ("Hydrazine", Chem.MolFromSmarts("[$([NH2]N)]"), 8.1, 1),  # Wikiped
    (
        "N-acetyl Piperazine",
        Chem.MolFromSmarts("[N+&!H0,NX3&H0;$(N1(C)CCN(C(=O)C)CC1)]"),
        7.1,
        1,
    ),  # Reaxys
    ("Piperazine", Chem.MolFromSmarts("C1CNCCN1"), 9.73, 1),  # Wikiped
    ("Pyridine", Chem.MolFromSmarts("n1ccccc1"), 5.25, 1),  # Wikiped
    ("Sulfonate", Chem.MolFromSmarts("[$([OH]S(=O)=O)]"), -1.5, 0),  # Guess
    (
        "Phosphonate",
        Chem.MolFromSmarts("[OH]P([OH])(=O)[#6]"),
        [1.5, 7.0],
        [0, -1],
    ),  # Guess
    (
        "7H-Pyrimido[4,5-b][1,4]oxazine-2,4-diamine",
        Chem.MolFromSmarts("C1C=NC2=CN=C(N=C2O1)N"),
        3.6,
        1,
    ),  # PKA Guessed
    ("Tyrosine", Chem.MolFromSmarts("[$([O-,OH]c1ccccc1)]"), 10.3, 0),
    (
        "Histidine",
        Chem.MolFromSmarts(
            "[CH2X4][#6X3]1:[$([#7X3H+,#7X2H0+0]:[#6X3H]:[#7X3H]),$([#7X3H])]:[#6X3H]:[$([#7X3H+,#7X2H0+0]:[#6X3H]:[#7X3H]),$([#7X3H])]:[#6X3H]1"
        ),
        6.6,
        1,
    ),
    ("Histidine1", Chem.MolFromSmarts("[$(cC)]1cnc[nH]1"), 6.6, 1),
    ("Histidine2", Chem.MolFromSmarts("[$(cC)]1c[nH]cn1"), 6.6, 1),
    (
        "Glutamic Acid",
        Chem.MolFromSmarts("[$([CX3][CD2][CD2]C(N)C(=O))](=O)[OX1H0-,OX2H1]"),
        4.2,
        0,
    ),
    ("Cysteine", Chem.MolFromSmarts("[$([S-,SH][CD2]C(N)C(=O))]"), 6.8, 0),
    (
        "Condensed Cysteine",
        Chem.MolFromSmarts("[$([SH][#175])]"),
        6.8,
        0,
    ),  # Sidechain treated as normal atom
    (
        "Condensed Cysteine",
        Chem.MolFromSmarts("[$([SH][#178])]"),
        6.8,
        0,
    ),  # Sidechain treated as normal atom
    # ('Aspartic Acid', Chem.MolFromSmarts('[$([CX3][CD2])](=O)[OX1H0-,OX2H1]'),3.90,0),
    (
        "Aspartic Acid v2",
        Chem.MolFromSmarts("[$([CX3][CD2]C(N)C(=O))](=O)[OX1H0-,OX2H1]"),
        3.5,
        0,
    ),
    ("Arginine", Chem.MolFromSmarts("[$(N[CD2][CD2])]C(=N)N"), 12.0, 1),
    ("Lysine", Chem.MolFromSmarts("[$([ND1][CD2][CD2][CD2][CD2]C(N)C(=O))]"), 10.5, 1),
    (
        "End Carboxylate (avg)",
        Chem.MolFromSmarts("[$([CX3]CN)](=O)[OX1H0-,OX2H1]"),
        3.3,
        0,
    ),
    # ('End Amine (avg)',Chem.MolFromSmarts('[$([NX3;H2,H1;!$(NC=O)]CC=O)]'),9.15,1),
    (
        "End Amine (avg) PseudoAtom safe",
        Chem.MolFromSmarts(
            "[$([NX3;H2,H1;!$(NC=O);!$(N[#171,#172,#173,#174,#175,#176,#177,#178,#179,#180,#181,#182,#183,#184,#185,#186,#187,#188,#189,#190,#191,#192,#193])]CC=O)]"
        ),
        7.7,
        1,
    ),
    ("Acrydine", Chem.MolFromSmarts("c1c2ccccc2nc2ccccc21"), 6.15, 1),
    ("Generic Carboxylic Acid", Chem.MolFromSmarts("[CX3](=O)[OX1H0-,OX2H1]"), 4.75, 0),
    (
        "Generic tert. Amine, not amide",
        Chem.MolFromSmarts("[N+&!H0&NX3,NX3;$(N(C)(C)C);!$(NC=[!#6]);!$(NC#[!#6])]"),
        9.8,
        1,
    ),
    (
        "Generic Amine, not amide",
        Chem.MolFromSmarts(
            "[N+,NX3;!H0;!$(NC=[!#6]);!$(NC#[!#6])!$(N[#171,#172,#173,#174,#175,#176,#177,#178,#179,#180,#181,#182,#183,#184,#185,#186,#187,#188,#189,#190,#191,#192,#193])]"
        ),
        10.0,
        1,
    ),
]

condensed_grimsley = [
    ("Condensed Tyrosine", Chem.MolFromSmarts("[#189]"), 10.3, 0),
    ("Condensed Histidine", Chem.MolFromSmarts("[#179]"), 6.6, 1),
    ("Condensed Glutamic Acid", Chem.MolFromSmarts("[#177]"), 4.2, 0),
    ("Condensed Aspartic Acid", Chem.MolFromSmarts("[#174]"), 3.5, 0),
    ("Condensed Arginine", Chem.MolFromSmarts("[#172]"), 12.0, 1),
    ("Condensed Lysine", Chem.MolFromSmarts("[#182]"), 10.5, 1),
    # ('Condensed AA end amino subst.', Chem.MolFromSmarts('[#0D1]'),9.15,1),
    # Can it somehow match up the wrong end? with a protection group??
    (
        "Condensed AA end amino",
        Chem.MolFromSmarts(
            "[#171,#172,#173,#174,#176,#177,#178,#179,#180,#181,#182,#183,#184,#185,#186,#187,#188,#189,#190,#191,#192,#193;D1]"
        ),
        7.7,
        1,
    ),
    # ('Condensed AA end Carboxylic Subst.', Chem.MolFromSmarts('[$([#0][OH])]'),2.15,0), #All acidic/basic should be replaced with glycine so they can be removed??
    ("Condensed cys end amino", Chem.MolFromSmarts("[#175;D2]"), 7.7, 1),
    # ('Condensed AA end Carboxylic Subst.', Chem.MolFromSmarts('[$([#0][OH])]'),2.15,0), #All acidic/basic should be replaced with glycine so they can be removed??
    (
        "Condensed AA end Carboxylic",
        Chem.MolFromSmarts(
            "[$([#171,#172,#173,#174,#175,#176,#177,#178,#179,#180,#181,#182,#183,#184,#185,#186,#187,#188,#189,#190,#191,#192,#193][OH])]"
        ),
        3.3,
        0,
    ),
]

# IPC_peptide as refered in KOZLOWSKI, Lukasz P. "IPC - Isoelectric Point Calculator.". Biol. Direct. 2016, vol 11, p. 55.
pkatable_ipc = [
    (
        "Tetramethylrhodamine #3",
        Chem.MolFromSmiles(
            "CN(C)C1=C-C=C2-C(C4=C(C(=O)O)C=CC=C4)=C3-C=C-C(=[N+](C)C)C=C3OC2=C1"
        ),
        [-10, 1.19, 3.59, 10],
        [1, 1, 0, 1],
    ),
    # Guess DOI: 10.1039/C0OB01045F
    (
        "6-hydroxy-3H-xanthen-3-one",
        Chem.MolFromSmarts("c2c3ccc(O)cc3oc3cc(=O)ccc2-3"),
        [2.94, 6.56],
        [1, 0],
    ),
    ("Anilin", Chem.MolFromSmarts("[$([NH2]c1ccccc1)]"), 4.87, 1),
    ("N-acetyl Hydrazine", Chem.MolFromSmarts("[$([NH2]NC(=O))]"), 2.81, 1),  # Reaxys
    ("Hydrazine", Chem.MolFromSmarts("[$([NH2]N)]"), 8.1, 1),  # Wikiped
    (
        "N-acetyl Piperazine",
        Chem.MolFromSmarts("[N+&!H0,NX3&H0;$(N1(C)CCN(C(=O)C)CC1)]"),
        7.1,
        1,
    ),  # Reaxys
    ("Piperazine", Chem.MolFromSmarts("C1CNCCN1"), 9.73, 1),  # Wikiped
    ("Pyridine", Chem.MolFromSmarts("n1ccccc1"), 5.25, 1),  # Wikiped
    ("Sulfonate", Chem.MolFromSmarts("[$([OH]S(=O)=O)]"), -1.5, 0),  # Guess
    (
        "Phosphonate",
        Chem.MolFromSmarts("[OH]P([OH])(=O)[#6]"),
        [1.5, 7.0],
        [0, -1],
    ),  # Guess
    (
        "7H-Pyrimido[4,5-b][1,4]oxazine-2,4-diamine",
        Chem.MolFromSmarts("C1C=NC2=CN=C(N=C2O1)N"),
        3.6,
        1,
    ),  # PKA Guessed
    ("Tyrosine", Chem.MolFromSmarts("[$([O-,OH]c1ccccc1)]"), 10.071, 0),
    (
        "Histidine",
        Chem.MolFromSmarts(
            "[CH2X4][#6X3]1:[$([#7X3H+,#7X2H0+0]:[#6X3H]:[#7X3H]),$([#7X3H])]:[#6X3H]:[$([#7X3H+,#7X2H0+0]:[#6X3H]:[#7X3H]),$([#7X3H])]:[#6X3H]1"
        ),
        6.018,
        1,
    ),
    ("Histidine1", Chem.MolFromSmarts("[$(cC)]1cnc[nH]1"), 6.018, 1),
    ("Histidine2", Chem.MolFromSmarts("[$(cC)]1c[nH]cn1"), 6.018, 1),
    (
        "Glutamic Acid",
        Chem.MolFromSmarts("[$([CX3][CD2][CD2]C(N)C(=O))](=O)[OX1H0-,OX2H1]"),
        4.317,
        0,
    ),
    ("Cysteine", Chem.MolFromSmarts("[$([S-,SH][CD2]C(N)C(=O))]"), 8.297, 0),
    (
        "Condensed Cysteine",
        Chem.MolFromSmarts("[$([SH][#175])]"),
        8.297,
        0,
    ),  # Sidechain treated as normal atom
    (
        "Condensed Cysteine",
        Chem.MolFromSmarts("[$([SH][#178])]"),
        8.297,
        0,
    ),  # Sidechain treated as normal atom
    # ('Aspartic Acid', Chem.MolFromSmarts('[$([CX3][CD2])](=O)[OX1H0-,OX2H1]'),3.90,0),
    (
        "Aspartic Acid v2",
        Chem.MolFromSmarts("[$([CX3][CD2]C(N)C(=O))](=O)[OX1H0-,OX2H1]"),
        3.887,
        0,
    ),
    ("Arginine", Chem.MolFromSmarts("[$(N[CD2][CD2])]C(=N)N"), 12.503, 1),
    (
        "Lysine",
        Chem.MolFromSmarts("[$([ND1][CD2][CD2][CD2][CD2]C(N)C(=O))]"),
        10.517,
        1,
    ),
    (
        "End Carboxylate (avg)",
        Chem.MolFromSmarts("[$([CX3]CN)](=O)[OX1H0-,OX2H1]"),
        2.383,
        0,
    ),
    # ('End Amine (avg)',Chem.MolFromSmarts('[$([NX3;H2,H1;!$(NC=O)]CC=O)]'),9.15,1),
    (
        "End Amine (avg) PseudoAtom safe",
        Chem.MolFromSmarts(
            "[$([NX3;H2,H1;!$(NC=O);!$(N[#171,#172,#173,#174,#175,#176,#177,#178,#179,#180,#181,#182,#183,#184,#185,#186,#187,#188,#189,#190,#191,#192,#193])]CC=O)]"
        ),
        9.564,
        1,
    ),
    ("Acrydine", Chem.MolFromSmarts("c1c2ccccc2nc2ccccc21"), 6.15, 1),
    ("Generic Carboxylic Acid", Chem.MolFromSmarts("[CX3](=O)[OX1H0-,OX2H1]"), 4.75, 0),
    (
        "Generic tert. Amine, not amide",
        Chem.MolFromSmarts("[N+&!H0&NX3,NX3;$(N(C)(C)C);!$(NC=[!#6]);!$(NC#[!#6])]"),
        9.8,
        1,
    ),
    (
        "Generic Amine, not amide",
        Chem.MolFromSmarts(
            "[N+,NX3;!H0;!$(NC=[!#6]);!$(NC#[!#6])!$(N[#171,#172,#173,#174,#175,#176,#177,#178,#179,#180,#181,#182,#183,#184,#185,#186,#187,#188,#189,#190,#191,#192,#193])]"
        ),
        10.0,
        1,
    ),
]

condensed_ipc = [
    ("Condensed Tyrosine", Chem.MolFromSmarts("[#189]"), 10.071, 0),
    ("Condensed Histidine", Chem.MolFromSmarts("[#179]"), 6.018, 1),
    ("Condensed Glutamic Acid", Chem.MolFromSmarts("[#177]"), 4.317, 0),
    ("Condensed Aspartic Acid", Chem.MolFromSmarts("[#174]"), 3.887, 0),
    ("Condensed Arginine", Chem.MolFromSmarts("[#172]"), 12.503, 1),
    ("Condensed Lysine", Chem.MolFromSmarts("[#182]"), 10.517, 1),
    # ('Condensed AA end amino subst.', Chem.MolFromSmarts('[#0D1]'),9.15,1),
    # Can it somehow match up the wrong end? with a protection group??
    (
        "Condensed AA end amino",
        Chem.MolFromSmarts(
            "[#171,#172,#173,#174,#176,#177,#178,#179,#180,#181,#182,#183,#184,#185,#186,#187,#188,#189,#190,#191,#192,#193;D1]"
        ),
        9.564,
        1,
    ),
    # ('Condensed AA end Carboxylic Subst.', Chem.MolFromSmarts('[$([#0][OH])]'),2.15,0), #All acidic/basic should be replaced with glycine so they can be removed??
    ("Condensed cys end amino", Chem.MolFromSmarts("[#175;D2]"), 9.564, 1),
    # ('Condensed AA end Carboxylic Subst.', Chem.MolFromSmarts('[$([#0][OH])]'),2.15,0), #All acidic/basic should be replaced with glycine so they can be removed??
    (
        "Condensed AA end Carboxylic",
        Chem.MolFromSmarts(
            "[$([#171,#172,#173,#174,#175,#176,#177,#178,#179,#180,#181,#182,#183,#184,#185,#186,#187,#188,#189,#190,#191,#192,#193][OH])]"
        ),
        2.383,
        0,
    ),
]

# EMBOSS iep. 2017, http://emboss.sourceforge.net/apps/release/6.6/emboss/apps/iep.html.
pkatable_EMBOSS = [
    (
        "Tetramethylrhodamine #3",
        Chem.MolFromSmiles(
            "CN(C)C1=C-C=C2-C(C4=C(C(=O)O)C=CC=C4)=C3-C=C-C(=[N+](C)C)C=C3OC2=C1"
        ),
        [-10, 1.19, 3.59, 10],
        [1, 1, 0, 1],
    ),
    # Guess DOI: 10.1039/C0OB01045F
    (
        "6-hydroxy-3H-xanthen-3-one",
        Chem.MolFromSmarts("c2c3ccc(O)cc3oc3cc(=O)ccc2-3"),
        [2.94, 6.56],
        [1, 0],
    ),
    ("Anilin", Chem.MolFromSmarts("[$([NH2]c1ccccc1)]"), 4.87, 1),
    ("N-acetyl Hydrazine", Chem.MolFromSmarts("[$([NH2]NC(=O))]"), 2.81, 1),  # Reaxys
    ("Hydrazine", Chem.MolFromSmarts("[$([NH2]N)]"), 8.1, 1),  # Wikiped
    (
        "N-acetyl Piperazine",
        Chem.MolFromSmarts("[N+&!H0,NX3&H0;$(N1(C)CCN(C(=O)C)CC1)]"),
        7.1,
        1,
    ),  # Reaxys
    ("Piperazine", Chem.MolFromSmarts("C1CNCCN1"), 9.73, 1),  # Wikiped
    ("Pyridine", Chem.MolFromSmarts("n1ccccc1"), 5.25, 1),  # Wikiped
    ("Sulfonate", Chem.MolFromSmarts("[$([OH]S(=O)=O)]"), -1.5, 0),  # Guess
    (
        "Phosphonate",
        Chem.MolFromSmarts("[OH]P([OH])(=O)[#6]"),
        [1.5, 7.0],
        [0, -1],
    ),  # Guess
    (
        "7H-Pyrimido[4,5-b][1,4]oxazine-2,4-diamine",
        Chem.MolFromSmarts("C1C=NC2=CN=C(N=C2O1)N"),
        3.6,
        1,
    ),  # PKA Guessed
    ("Tyrosine", Chem.MolFromSmarts("[$([O-,OH]c1ccccc1)]"), 10.1, 0),
    (
        "Histidine",
        Chem.MolFromSmarts(
            "[CH2X4][#6X3]1:[$([#7X3H+,#7X2H0+0]:[#6X3H]:[#7X3H]),$([#7X3H])]:[#6X3H]:[$([#7X3H+,#7X2H0+0]:[#6X3H]:[#7X3H]),$([#7X3H])]:[#6X3H]1"
        ),
        6.5,
        1,
    ),
    ("Histidine1", Chem.MolFromSmarts("[$(cC)]1cnc[nH]1"), 6.5, 1),
    ("Histidine2", Chem.MolFromSmarts("[$(cC)]1c[nH]cn1"), 6.5, 1),
    (
        "Glutamic Acid",
        Chem.MolFromSmarts("[$([CX3][CD2][CD2]C(N)C(=O))](=O)[OX1H0-,OX2H1]"),
        4.1,
        0,
    ),
    ("Cysteine", Chem.MolFromSmarts("[$([S-,SH][CD2]C(N)C(=O))]"), 8.5, 0),
    (
        "Condensed Cysteine",
        Chem.MolFromSmarts("[$([SH][#175])]"),
        8.5,
        0,
    ),  # Sidechain treated as normal atom
    (
        "Condensed Cysteine",
        Chem.MolFromSmarts("[$([SH][#178])]"),
        8.5,
        0,
    ),  # Sidechain treated as normal atom
    # ('Aspartic Acid', Chem.MolFromSmarts('[$([CX3][CD2])](=O)[OX1H0-,OX2H1]'),3.90,0),
    (
        "Aspartic Acid v2",
        Chem.MolFromSmarts("[$([CX3][CD2]C(N)C(=O))](=O)[OX1H0-,OX2H1]"),
        3.9,
        0,
    ),
    ("Arginine", Chem.MolFromSmarts("[$(N[CD2][CD2])]C(=N)N"), 12.5, 1),
    ("Lysine", Chem.MolFromSmarts("[$([ND1][CD2][CD2][CD2][CD2]C(N)C(=O))]"), 10.8, 1),
    (
        "End Carboxylate (avg)",
        Chem.MolFromSmarts("[$([CX3]CN)](=O)[OX1H0-,OX2H1]"),
        3.6,
        0,
    ),
    # ('End Amine (avg)',Chem.MolFromSmarts('[$([NX3;H2,H1;!$(NC=O)]CC=O)]'),9.15,1),
    (
        "End Amine (avg) PseudoAtom safe",
        Chem.MolFromSmarts(
            "[$([NX3;H2,H1;!$(NC=O);!$(N[#171,#172,#173,#174,#175,#176,#177,#178,#179,#180,#181,#182,#183,#184,#185,#186,#187,#188,#189,#190,#191,#192,#193])]CC=O)]"
        ),
        8.6,
        1,
    ),
    ("Acrydine", Chem.MolFromSmarts("c1c2ccccc2nc2ccccc21"), 6.15, 1),
    ("Generic Carboxylic Acid", Chem.MolFromSmarts("[CX3](=O)[OX1H0-,OX2H1]"), 4.75, 0),
    (
        "Generic tert. Amine, not amide",
        Chem.MolFromSmarts("[N+&!H0&NX3,NX3;$(N(C)(C)C);!$(NC=[!#6]);!$(NC#[!#6])]"),
        9.8,
        1,
    ),
    (
        "Generic Amine, not amide",
        Chem.MolFromSmarts(
            "[N+,NX3;!H0;!$(NC=[!#6]);!$(NC#[!#6])!$(N[#171,#172,#173,#174,#175,#176,#177,#178,#179,#180,#181,#182,#183,#184,#185,#186,#187,#188,#189,#190,#191,#192,#193])]"
        ),
        10.0,
        1,
    ),
]

condensed_EMBOSS = [
    ("Condensed Tyrosine", Chem.MolFromSmarts("[#189]"), 10.1, 0),
    ("Condensed Histidine", Chem.MolFromSmarts("[#179]"), 6.5, 1),
    ("Condensed Glutamic Acid", Chem.MolFromSmarts("[#177]"), 4.4, 0),
    ("Condensed Aspartic Acid", Chem.MolFromSmarts("[#174]"), 3.9, 0),
    ("Condensed Arginine", Chem.MolFromSmarts("[#172]"), 12.5, 1),
    ("Condensed Lysine", Chem.MolFromSmarts("[#182]"), 10.8, 1),
    # ('Condensed AA end amino subst.', Chem.MolFromSmarts('[#0D1]'),9.15,1),
    # Can it somehow match up the wrong end? with a protection group??
    (
        "Condensed AA end amino",
        Chem.MolFromSmarts(
            "[#171,#172,#173,#174,#176,#177,#178,#179,#180,#181,#182,#183,#184,#185,#186,#187,#188,#189,#190,#191,#192,#193;D1]"
        ),
        8.6,
        1,
    ),
    # ('Condensed AA end Carboxylic Subst.', Chem.MolFromSmarts('[$([#0][OH])]'),2.15,0), #All acidic/basic should be replaced with glycine so they can be removed??
    ("Condensed cys end amino", Chem.MolFromSmarts("[#175;D2]"), 8.6, 1),
    # ('Condensed AA end Carboxylic Subst.', Chem.MolFromSmarts('[$([#0][OH])]'),2.15,0), #All acidic/basic should be replaced with glycine so they can be removed??
    (
        "Condensed AA end Carboxylic",
        Chem.MolFromSmarts(
            "[$([#171,#172,#173,#174,#175,#176,#177,#178,#179,#180,#181,#182,#183,#184,#185,#186,#187,#188,#189,#190,#191,#192,#193][OH])]"
        ),
        3.6,
        0,
    ),
]

# Defaut pKa_tables
pkatable = pkatable_grimsley
# pkatable = pkatable_ipc
# pkatable = pkatable_EMBOSS

condensedtable = condensed_grimsley


def find_pKas(
    mol,
    pkatable=pkatable,
    condensedtable=condensedtable,
    debug=False,
    returnindices=False,
):
    """This function finds pKa values in a supplied molecule by matching it up to SMARTS rules in a table
    Returns the list of pKa values and a list with the charge of the group at pH levels blow the pKa.
    Example: So histidine has a +1 charge at pH levels below the pKa of ~6"""
    chargelist = []
    pkalist = []
    dummy = Chem.MolFromSmarts(
        "[#178]"
    )  # Used Gly instead of #0 or *, for compatibility with building blocks

    if returnindices:
        atomindices = []
        for atom in mol.GetAtoms():
            atom.SetProp(
                "O_i", str(atom.GetIdx())
            )  # Save the Original Idx as a property

    # Look for condensed AA atoms with potential pKa values. Substitute with Glycine
    for s in condensedtable:
        if mol.HasSubstructMatch(s[1]):
            if debug:
                print
            "Match: %s" % (s[0]),
            # How many??
            substructmatches = mol.GetSubstructMatches(s[1])
            n = len(substructmatches)
            if debug:
                print
            "times %s" % (n)
            pkalist += [s[2]] * n
            chargelist += [s[3]] * n
            if returnindices:
                for match in substructmatches:
                    atomindices += [
                        [int(mol.GetAtomWithIdx(idx).GetProp("O_i")) for idx in match]
                    ]  # Retrieve the original index of matches
            mol = AllChem.ReplaceSubstructs(mol, s[1], dummy)[
                0
            ]  # Delete acidic/basic groups to prevent multiple matching

    # Look for molecular fragments with known pKa's
    for s in pkatable:
        # print s[0]
        if mol.HasSubstructMatch(s[1]):
            if debug:
                print
            "Match: %s" % (s[0]),
            # How many??
            substructmatches = mol.GetSubstructMatches(s[1])
            n = len(substructmatches)
            if debug:
                print
            "times %s" % (n)
            if type(s[2]) == type(float()):
                pkalist += [s[2]] * n
                chargelist += [s[3]] * n
            else:  # If there's multiple pKa values found for the molecular substructure
                for i in range(len(s[2])):
                    pkalist += [s[2][i]] * n
                    chargelist += [s[3][i]] * n
            if returnindices:
                for match in substructmatches:
                    atomindices += [
                        [int(mol.GetAtomWithIdx(idx).GetProp("O_i")) for idx in match]
                    ]  # Retrieve the original i
            mol = AllChem.ReplaceSubstructs(mol, s[1], dummy)[0]
    if returnindices:
        return pkalist, chargelist, atomindices
    return pkalist, chargelist


def charge(ph, pkalist, chargelist):
    """Jacob Tolborgs charge model where the charge is assigned from partial charges from all pKa values at the pH point"""
    chargesum = []
    abschargesum = []
    for charge, pka in zip(chargelist, pkalist):
        # print charge, pka
        if charge == 1:
            charge = 1 / (1 + 10 ** (ph - pka))
            chargesum.append(charge)
        else:
            charge = -1 / (1 + 10 ** -(ph - pka))
            chargesum.append(charge)
            abschargesum.append(abs(charge))
    return sum(chargesum), sum(abschargesum)


def pI(pkalist, chargelist):
    """Uses Jacob Tolborgs charge function and calculates pI.
    If only only acidic or basic groups are found, the pI is ill defined and returned as 42 or -42"""
    # Check if Only acidic or basic groups are present.
    if len(unique(chargelist)) == 0:  # No pKa groups present.
        pI = 7
    elif len(unique(chargelist)) == 1:  # Only one type is present
        if chargelist[0] == 0:  # Only acidic groups are present
            pI = 3
        elif chargelist[0] == 1:  # Only basic groups are present
            pI = 11
    else:
        # Find pI by simulation in the pH range 0 to 0
        chargecol = []
        for i in range(0, 1400):
            ph = i / 100.0
            chargecol.append(charge(ph, pkalist, chargelist))  # Calculate charge
        pI = (
            argmin(np_abs(chargecol)) / 100.0
        )  # Simply taking the smallest absolute value, and dividing the index with 100
    # print "pI %.1f"%pI
    return pI


import numpy as np

from olorenchemengine.base_class import log_arguments
from olorenchemengine.representations import BaseCompoundVecRepresentation


class calc_pI(BaseCompoundVecRepresentation):
    @log_arguments
    def __init__(self, ph=7.4, log=True, **kwargs):
        self.ph = ph
        super().__init__(
            names=["pI", f"Charge{self.ph}", f"TotalAbsCharge{self.ph}"],
            log=False,
            **kwargs,
        )

    def _convert(self, smiles, y=None):
        # find pKa values and charge class
        m = Chem.MolFromSmiles(smiles)
        pkalist, chargelist = find_pKas(m)
        # Calculate pI
        pIpred = pI(pkalist, chargelist)
        return np.array([pIpred, *charge(self.ph, pkalist, chargelist)])



"""Techniques for quantifying uncertainty and estimating confidence intervals for all oce models.
"""

import numpy as np
import pandas as pd
from rdkit.Chem import AllChem
from rdkit.DataStructs.cDataStructs import BulkTanimotoSimilarity

import olorenchemengine as oce

from .base_class import *
from .basics import *
from .dataset import *
from .representations import *
from .reduction import *


class BaseFingerprintModel(BaseErrorModel):
    """Morgan fingerprint-based error models.
    
    BaseFingerprintModel is the base class for error models that require the
    computation of Morgan fingerprints.
    """

    @log_arguments
    def __init__(self, radius=2, log=True, **kwargs):
        self.radius = radius      
        super().__init__(log=False, **kwargs)

    def _build(self, **kwargs):
        if isinstance(self.X_train, pd.DataFrame):
            self.X_train.columns = ["SMILES"]
        self.train_fps = list(self.get_fps(self.X_train))

    def get_fps(self, smiles: List[str]) -> List:
        smiles = np.array(SMILESRepresentation().convert(smiles))
        get_fp = lambda smi: AllChem.GetMorganFingerprint(Chem.MolFromSmiles(smi), self.radius)
        return list(np.vectorize(get_fp)(smiles))

    def _save(self) -> dict:
        d = super()._save()
        if hasattr(self, "train_fps"):
            d.update({"train_fps": self.train_fps})
        return d

    def _load(self, d) -> None:
        super()._load(d)
        if "train_fps" in d.keys():
            self.train_fps = d["train_fps"]

# List of kernels for kernel error models
KERNELS = {
    "sdc":          lambda X, h: np.exp(-h * X / (1 - X)),
    "uniform":      lambda X, h: h * X < 1,
    "linear":       lambda X, h: np.maximum(0, 1 - h * X),
    "parabolic":    lambda X, h: np.maximum(0, 1 - (h * X) ** 2),
    "power":        lambda X, h: (1 - X) ** h,
    "gaussian":     lambda X, h: np.exp(-(h * X) ** 2),
}

# List of predictors for kernel error models
PREDICTORS = {
    "property": lambda y, em: em.y_train - y,
    "error":    lambda y, em: em.y_train - em.y_pred_train,
}

class KernelError(BaseFingerprintModel):
    """Kernel error model. 
    
    KernelError uses a kernel-weighted average of prediction errors as the
    covariate for estimating confidence intervals. It is inspired by the 
    Nadaraya-Watson estimator, which generates a regression using a
    kernel-weighted average. The distance function used is 1 - Tanimoto
    Similarity.

    This is the recommended error model for general purposes and models.

    Parameters
    ------------------------------
    predictor (str, {"property", "error"}): Error predictor being estimated
    kernel (str, {"default"}): Kernel used as a weight-function
    h (int, float): Bandwidth

    Example
    ------------------------------
    import olorenchemengine as oce
    model = oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000)
    model.fit_cv(train["Drug"], train["Y"], error_model = oce.KernelError())
    model.predict(test["Drug"], return_ci = True)
    """

    @log_arguments
    def __init__(self, predictor="property", kernel="sdc", h=3, log=True, **kwargs):
        assert predictor in PREDICTORS, "Predictor `{}` is invalid. Please choose one of `{}`.".format(predictor, "`, `".join(KERNELS.keys()))
        assert kernel in KERNELS, "Kernel `{}` is invalid. Please choose one of `{}`.".format(kernel, "`, `".join(KERNELS.keys()))

        self.predictor = predictor
        self.kernel = kernel
        self.h = h
        super().__init__(log=False, **kwargs)

    def calculate(self, X, y_pred):
        K = KERNELS[self.kernel]
        P = PREDICTORS[self.predictor]

        def covariate(smiles, y):
            mol = Chem.MolFromSmiles(smiles)
            fp = AllChem.GetMorganFingerprint(mol, 2)
            TD = 1 - np.array(BulkTanimotoSimilarity(fp, self.train_fps))
            w = K(TD, self.h)
            p = P(y, self)

            norm = np.sum(w)
            if norm == 0:
                return np.mean(p)
            else:
                return np.dot(w, p) / norm

        X = np.array(X).flatten()
        return np.array([covariate(smiles, y) for smiles, y in tqdm(zip(X, y_pred))])


class SDC(BaseFingerprintModel):
    """ SDC is an error model that predicts error bars based on the Sum of
        Distance-weighted Contributions: `Molecular Similarity-Based Domain
        Applicability Metric Efficiently Identifies Out-of-Domain Compounds
        <http://dx.doi.org/10.1021/acs.jcim.8b00597>`_

        Parameters:
            a (int or float, optional): Value of a in the SDC formula. Default 3.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000)
    model.fit_cv(train["Drug"], train["Y"], error_model = oce.SDC())
    model.predict(test["Drug"], return_ci = True)
    ------------------------------
    """

    @log_arguments
    def __init__(self, a=3, log=True, **kwargs):
        self.a = a
        super().__init__(log=False, **kwargs)

    def calculate(self, X, y_pred):
        ref_fps = self.get_fps(X)

        def sdc(fp):
            TS = np.array(BulkTanimotoSimilarity(fp, self.train_fps))
            return np.sum(np.exp(-self.a * (1 - TS) / TS))

        return np.array([sdc(fp) for fp in tqdm(ref_fps)])


class TargetDistDC(SDC):
    """ TargetDistDC is an error model that calculates the root-mean-square
        difference between the predicted activity of the target molecule and
        the observed activities of all training molecules, weighted by the DC.

        Parameters:
            a (int or float, optional): Value of a in the SDC formula. Default 3.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000)
    model.fit_cv(train["Drug"], train["Y"], error_model = oce.TargetDistDC())
    model.predict(test["Drug"], return_ci = True)
    ------------------------------
    """

    @log_arguments
    def __init__(self, log=True, **kwargs):      
        super().__init__(log=False, **kwargs)

    def calculate(self, X, y_pred):        
        def dist(smi, pred):
            mol = Chem.MolFromSmiles(smi)
            ref_fp = AllChem.GetMorganFingerprint(mol, 2)
            TS = np.array(BulkTanimotoSimilarity(ref_fp, self.train_fps))
            DC = np.exp(-self.a * (1 - TS) / TS)
            error = np.abs(self.y_train - pred)
            if np.sum(DC) == 0:
                return np.sqrt(np.mean(error))
            return np.sqrt(np.dot(DC, error) / np.sum(DC))

        X = np.array(X).flatten()
        return np.array([dist(smi, pred) for smi, pred in tqdm(zip(X, y_pred))])


class TrainDistDC(SDC):
    """ TrainDistDC is an error model that calculates the root-mean-square
        difference between the predicted and observed activities of all
        training molecules, weighted by the DC.

        Parameters:
            a (int or float, optional): Value of a in the SDC formula. Default 3.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000)
    model.fit_cv(train["Drug"], train["Y"], error_model = oce.TrainDistDC())
    model.predict(test["Drug"], return_ci = True)
    ------------------------------
    """

    @log_arguments
    def __init__(self, log=True, **kwargs):      
        super().__init__(log=False, **kwargs)

    def calculate(self, X, y_pred):
        residuals = np.abs(self.y_train - self.y_pred_train)
        ref_fps = self.get_fps(X)

        def dist(fp):
            TS = np.array(BulkTanimotoSimilarity(fp, self.train_fps))
            DC = np.exp(-self.a * (1 - TS) / TS)
            if np.sum(DC) == 0:
                return np.sqrt(np.mean(residuals))
            return np.sqrt(np.dot(DC, residuals) / np.sum(DC))

        return np.array([dist(fp) for fp in tqdm(ref_fps)])


class KNNSimilarity(BaseFingerprintModel):
    """ NNSimilarity is an error model that calculates mean Tanimoto similarity
        between the target molecule and the k most similar training molecules
        using a Morgan Fingerprint with a radius of 2 bonds.

        Parameters:
            k (int, optional): Number of nearest neighbors to consider. Default 5.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000)
    model.fit_cv(train["Drug"], train["Y"], error_model = oce.KNNSimilarity())
    model.predict(test["Drug"], return_ci = True)
    ------------------------------
    """

    @log_arguments
    def __init__(self, k=5, log=True, **kwargs):
        self.k = k
        super().__init__(log=False, **kwargs)

    def calculate(self, X, y_pred):
        ref_fps = self.get_fps(X)

        def mean_sim(fp):
            similarity = np.array(BulkTanimotoSimilarity(fp, self.train_fps))
            return np.mean(np.partition(similarity, self.k)[-self.k:])

        return np.array([mean_sim(fp) for fp in tqdm(ref_fps)])

class TargetDistKNN(KNNSimilarity):
    """ TargetDistKNN is an error model that calculates the root-mean-square
        difference between the predicted activity of the target molecule and
        the observed activities of the k most similar training molecules,
        weighted by their similarity.

        Parameters:
            k (int, optional): Number of nearest neighbors to consider. Default 5.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000)
    model.fit_cv(train["Drug"], train["Y"], error_model = oce.TargetDistKNN())
    model.predict(test["Drug"], return_ci = True)
    ------------------------------
    """

    @log_arguments
    def __init__(self, log=True, **kwargs):      
        super().__init__(log=False, **kwargs)

    def calculate(self, X, y_pred):
        def dist(smi, pred):
            mol = Chem.MolFromSmiles(smi)
            ref_fp = AllChem.GetMorganFingerprint(mol, 2)
            TS = np.array(BulkTanimotoSimilarity(ref_fp, self.train_fps))
            error = np.abs(self.y_train - pred)
            idxs = np.argpartition(TS, self.k)[-self.k:]
            if np.sum(TS[idxs]) == 0:
                return np.sqrt(np.mean(error[idxs]))
            return np.sqrt(np.dot(TS[idxs], error[idxs]) / np.sum(TS[idxs]))

        X = np.array(X).flatten()
        return np.array([dist(smi, pred) for smi, pred in tqdm(zip(X, y_pred))])


class TrainDistKNN(KNNSimilarity):
    """ TrainDistKNN is an error model that calculates the root-mean-square
        difference between the predicted and observed activities of the k most
        similar training molecules to the target molecule, weighted by their
        similarity.

        Parameters:
            k (int, optional): Number of nearest neighbors to consider. Default 5.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000)
    model.fit_cv(train["Drug"], train["Y"], error_model = oce.TrainDistKNN())
    model.predict(test["Drug"], return_ci = True)
    ------------------------------
    """

    @log_arguments
    def __init__(self, log=True, **kwargs):      
        super().__init__(log=False, **kwargs)

    def calculate(self, X, y_pred):
        residuals = np.abs(self.y_train - self.y_pred_train)
        ref_fps = self.get_fps(X)

        def dist(fp):
            similarity = np.array(BulkTanimotoSimilarity(fp, self.train_fps))
            idxs = np.argpartition(similarity, self.k)[-self.k:]
            if np.sum(similarity[idxs]) == 0:
                return np.sqrt(np.mean(residuals[idxs]))
            return np.sqrt(np.dot(similarity[idxs], residuals[idxs]) / np.sum(similarity[idxs]))
        
        return np.array([dist(fp) for fp in tqdm(ref_fps)])


class Predicted(BaseErrorModel):
    """ Predicted is an error model that predicts error bars based on only the
        predicted value of a molecule. It is best used as part of an aggregate
        error model rather than by itself.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000)
    model.fit_cv(train["Drug"], train["Y"], error_model = oce.AggregateErrorModel([oce.SDC(), oce.Predicted()])
    model.predict(test["Drug"], return_ci = True)
    ------------------------------
    """

    @log_arguments
    def __init__(self, log=True, **kwargs):      
        super().__init__(log=False, **kwargs)

    def calculate(self, X, y_pred):
        return y_pred

class Naive(BaseErrorModel):
    """ Naive is an error model that predicts a uniform confidence interval
        based on the errors of the fitting dataset. Used exclusively for
        benchmarking error models.
    """

    @log_arguments
    def __init__(self, log=True, **kwargs):
        super().__init__(log=False, **kwargs)

    def calculate(self, X, y_pred):
        return np.array([0 for _ in X])
    
    def _fit(
        self,
        residuals: np.ndarray,
        scores: np.ndarray,
    ):
        self.error = np.quantile(residuals, self.ci)
        self.reg = lambda X: np.array([self.error for _ in X])

    def score(
        self,
        X: Union[pd.DataFrame, np.ndarray, list, pd.Series],
    ):
        return self.reg(X)

class BaseEnsembleModel(BaseErrorModel):
    """ BaseEnsembleModel is the base class for error models that estimate
        uncertainty based on the variance of an ensemble of models.
    """

    @log_arguments
    def __init__(self, ensemble_model = None, n_ensembles = 16, log=True, **kwargs):      
        self.ensemble_model = ensemble_model
        self.n_ensembles = n_ensembles
        super().__init__(log=False, **kwargs)

    def _build(self, **kwargs):
        if self.ensemble_model is None:
            self.ensemble_model = self.model.copy()
        self.ensembles = []
    
    def calculate(self, X, y_pred):
        predictions = np.stack([model.predict(X) for model in tqdm(self.ensembles)])
        return np.var(predictions, axis=0)

    def _save(self) -> dict:
        d = super()._save()
        if hasattr(self, "ensembles"):
            d.update({"ensembles": [saves(model) for model in self.ensembles]})
        return d

    def _load(self, d) -> None:
        super()._load(d)
        if "ensembles" in d.keys():
            self.ensembles = [loads(model) for model in d["ensembles"]]
    
class BootstrapEnsemble(BaseEnsembleModel):
    """ BootstrapEnsemble estimates uncertainty based on the variance of several
        models trained on bootstrapped samples of the training data.

        Parameters:
            ensemble_model (BaseModel): Model used for ensembling. Defaults to the same as the original model.
            n_ensembles (int): Number of ensembles
            bootstrap_size (float): Proportion of training data to train each ensemble model
    
    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000)
    model.fit_cv(train["Drug"], train["Y"], error_model = oce.BootstrapEnsemble(n_ensembles = 10))
    model.predict(test["Drug"], return_ci = True)
    ------------------------------
    """
    
    @log_arguments
    def __init__(self, ensemble_model = None, n_ensembles = 12, bootstrap_size = 0.25, log=True, **kwargs):
        self.bootstrap_size = bootstrap_size
        super().__init__(ensemble_model = ensemble_model, n_ensembles = n_ensembles, log=False, **kwargs)

    def _build(self, **kwargs):
        super()._build()

        from sklearn.model_selection import train_test_split
        
        for _ in tqdm(range(self.n_ensembles)):
            X_train, X_test, y_train, y_test = train_test_split(
                self.X_train, self.y_train, train_size=self.bootstrap_size
            )
            ensemble_model = self.ensemble_model.copy()
            ensemble_model.fit(X_train, y_train)
            self.ensembles.append(ensemble_model)

class RandomForestEnsemble(BaseEnsembleModel):
    """ RandomForestEnsemble estimates uncertainty based on the variance of several
        random forest models initialized to different random states.

        Parameters:
            ensemble_model (BaseModel): Model used for ensembling. Defaults to the same as the original model.
            n_ensembles (int): Number of ensembles
    
    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000)
    model.fit_cv(train["Drug"], train["Y"], error_model = oce.RandomForestEnsemble(n_ensembles = 10))
    model.predict(test["Drug"], return_ci = True)
    ------------------------------
    """

    @log_arguments
    def __init__(self, log=True, **kwargs):      
        super().__init__(log=False, **kwargs)
    
    def _build(self, **kwargs):
        super()._build()

        for n in tqdm(range(self.n_ensembles)):
            ensemble_model = RandomForestModel(
                oce.MorganVecRepresentation(radius=2, nbits=2048), 
                random_state=n, 
                **kwargs
            )
            ensemble_model.fit(self.X_train, self.y_train)
            self.ensembles.append(ensemble_model)

class ADAN(BaseErrorModel):
    """ ADAN is an error model that predicts error bars based on one or
        multiple ADAN categories: `Applicability Domain Analysis (ADAN): A
        Robust Method for Assessing the Reliability of Drug Property Predictions
        <https://doi.org/10.1021/ci500172z>`_

        Parameters:
            criterion (str): the ADAN criteria to be considered.
            rep (BaseCompoundVecRepresentation): the representation to use. By default, 
                usees the representation of the BaseModel object.
            dim_reduction ({"pls", "pca"}): the dimensionality reduction to use. 
            explvar (float): the desired variance to be captured by the dimensionality
                reduction components as a proportion of total variance.
            threshold (float): the quantile for a criterion to be considered as out of
                its standard range.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000)
    model.fit_cv(train["Drug"], train["Y"], error_model = oce.ADAN("E_raw"))
    model.predict(test["Drug"], return_ci = True)
    ------------------------------
    """

    @log_arguments
    def __init__(
        self, 
        criterion: str = "Category", 
        rep: BaseCompoundVecRepresentation = None,
        dim_reduction: str = "pls",
        explvar: float = 0.8,
        threshold: float = 0.95,
        log=True, 
        **kwargs
    ):
        self.criterion = criterion
        self.rep = rep
        self.dim_reduction = dim_reduction
        self.explvar = explvar
        self.threshold = threshold
        super().__init__(log=False, **kwargs)

    def _build(self, **kwargs):
        self.X_rep = self.preprocess(self.X_train)
        self.n_components = min(self.X_rep.shape)

        if self.dim_reduction == "pls":
            from sklearn.cross_decomposition import PLSRegression

            self.reduction = PLSRegression(n_components=self.n_components)
            self.reduction.fit(self.X_rep, self.y_train)
            x_var = np.var(self.reduction.x_scores_, axis=0)
            x_var /= np.sum(x_var)
        elif self.dim_reduction == "pca":
            from sklearn.decomposition import PCA

            self.reduction = PCA(n_components=self.n_components)
            self.reduction.fit(self.X_rep, self.y_train)
            x_var = self.reduction.explained_variance_ratio_
        else:
            raise NameError("dim_reduction {} is not recognized. Valid inputs are 'pls' and 'pca'.".format(self.dim_reduction))

        if np.sum(x_var) > self.explvar:
            self.n_components = np.where(np.cumsum(x_var) > self.explvar)[0][0] + 1

        self.Xp_train = self.reduction.transform(self.X_rep)[:, : self.n_components]
        self.Xp_mean = np.mean(self.Xp_train, axis=0)
        self.y_mean = np.mean(self.y_train)

        from sklearn.neighbors import NearestNeighbors

        nbrs = NearestNeighbors(n_neighbors=2).fit(self.Xp_train)
        distances, indices = nbrs.kneighbors(self.Xp_train)

        centroid_dist = np.linalg.norm(self.Xp_train - self.Xp_mean, axis=1)
        neighbor_dist = distances[:, 1]
        model_dist = self.DModX(self.X_rep, self.Xp_train)
        y_mean_dist = np.abs(self.y_train - self.y_mean)
        y_nei_dist = np.abs(self.y_train - self.y_train[indices[:, 1]])
        SDEP_dist = self.SDEP(self.Xp_train, n_drop=1)

        self.training_ = {
            "A_raw": centroid_dist,
            "B_raw": neighbor_dist,
            "C_raw": model_dist,
            "D_raw": y_mean_dist,
            "E_raw": y_nei_dist,
            "F_raw": SDEP_dist,
        }

        thresh = int(self.X_rep.shape[0] * self.threshold)
        self.thresholds_ = {
            "A": np.partition(centroid_dist, -thresh)[-thresh],
            "B": np.partition(neighbor_dist, -thresh)[-thresh],
            "C": np.partition(model_dist, -thresh)[-thresh],
            "D": np.partition(y_mean_dist, -thresh)[-thresh],
            "E": np.partition(y_nei_dist, -thresh)[-thresh],
            "F": np.partition(SDEP_dist, -thresh)[-thresh],
        }

    def calculate_full(self, X, standardize: bool = True):
        """Calculates complete confidence scores for visualization."""
        criteria = ["A", "B", "C", "D", "E", "F"]
        y_pred = np.array(self.model.predict(X)).flatten()
        X = self.preprocess(X)
        Xp = self.reduction.transform(X)[:, : self.n_components]

        self.results = {}
        for c in criteria:
            self.results[c] = self._calculate(X, Xp, y_pred, c)
        for c in criteria:
            c += "_raw"
            self.results[c] = self._calculate(X, Xp, y_pred, c, standardize=standardize)
        self.results["Category"] = np.sum([self.results[c] for c in criteria], axis=0)

        self.results = pd.DataFrame(self.results)

    def calculate(self, X, y_pred, standardize: bool = True):
        """Calcualtes confidence scores."""
        X = self.preprocess(X)
        Xp = self.reduction.transform(X)[:, : self.n_components]
        
        if self.criterion == "Category":
            criteria = ["A", "B", "C", "D", "E", "F"]
            return np.sum([self._calculate(X, Xp, y_pred, c) for c in criteria], axis=0)
        else:
            return self._calculate(X, Xp, y_pred, self.criterion, standardize=standardize)

    def _calculate(self, X, Xp, y_pred, criterion: str, standardize: bool = True):
        """Calcualtes confidence scores for a given criteron."""
        from sklearn.neighbors import NearestNeighbors
        
        if criterion in ("A", "A_raw"):
            dist = np.linalg.norm(Xp - self.Xp_mean, axis=1)
        elif criterion in ("B", "B_raw"):
            nbrs = NearestNeighbors(n_neighbors=1).fit(self.Xp_train)
            distances = nbrs.kneighbors(Xp)[0]
            dist = distances.flatten()
        elif criterion in ("C", "C_raw"):
            dist = self.DModX(X, Xp)
        elif criterion in ("D", "D_raw"):
            dist = self.SDEP(Xp, 0)
        elif criterion in ("E", "E_raw"):
            dist = np.abs(y_pred - self.y_mean)
        elif criterion in ("F", "F_raw"):
            nbrs = NearestNeighbors(n_neighbors=1).fit(self.Xp_train)
            indices = nbrs.kneighbors(Xp)[1]
            dist = np.abs(y_pred - self.y_train[indices.flatten()])
        else:
            raise NameError("Criterion {} is not recognized.".format(criterion))

        if len(criterion) > 1:
            if standardize:
                return (dist - np.mean(self.training_[criterion])) / np.std(
                    self.training_[criterion]
                )
            else:
                return dist
        else:
            return np.where(dist > self.thresholds_[criterion], 1, 0)

    def preprocess(self, X, y=None):
        """Preprocesses data into the appropriate representation."""
        if self.rep is None:
            X = self.model.preprocess(X, y)
        else:
            X = np.array(self.rep.convert(X))

        assert isinstance(
            X, np.ndarray
        ), "The preprocess for the model must return a np.ndarray, e.g. the model must use a BaseCompoundVecRepresentation or a representation must be passed"

        return X

    def DModX(self, X: np.ndarray, Xp: np.ndarray) -> np.ndarray:
        """Computes the distance to the model (DmodX).

        Computes the distance between a datapoint and the PLS model plane. See
        <https://www.jmp.com/support/help/en/15.2/index.shtml#page/jmp/dmodx-calculation.shtml>
        for more details about the statistic.

        Parameters:
            X (np.ndarray): queries
            Xp (np.ndarray): queries transformed into latent space
        """
        if self.dim_reduction == "pls":
            X_reconstructed = (
                Xp
                @ self.reduction.x_loadings_[:, : self.n_components].T
                * self.reduction._x_std
                + self.reduction._x_mean
            )
        elif self.dim_reduction == "pca":
            X_reconstructed = (
                np.dot(Xp, self.reduction.components_[: self.n_components, :])
                + self.reduction.mean_
            )
        return np.linalg.norm(X - X_reconstructed, axis=1) / np.sqrt(
            self.X_rep.shape[1] - self.n_components
        )

    def SDEP(
        self, Xp: np.ndarray, n_drop: int = 0, neighbor_thresh: float = 0.05
    ) -> np.ndarray:
        """Computes the standard deviation error of predictions (SDEP).

        Computes the standard deviation training error of the `neighbor_thresh`
        fraction of closest training queries to each query in `Xp` in latent space.

        Parameters:
            Xp (np.ndarray): queries transformed into latent space
            n_drop (int): 1 if fitting, 0 if scoring
            neighbor_thresh (float): fraction of closest training queries to consider
        """
        from sklearn.neighbors import NearestNeighbors
        
        n_neighbors = int(self.X_rep.shape[0] * neighbor_thresh) + n_drop
        nbrs = NearestNeighbors(n_neighbors=n_neighbors).fit(self.Xp_train)
        distances, indices = nbrs.kneighbors(Xp)

        y_sqerr = np.array((self.y_train - self.y_pred_train) ** 2)
        y_mse = np.mean(y_sqerr[indices[:, n_drop:]], axis=1).astype(float)

        return np.sqrt(y_mse).flatten()

class AggregateErrorModel(BaseErrorModel):
    """ AggregateErrorModel estimates uncertainty by aggregating ucertainty scores from
        several different BaseErrorModels.

        Parameters:
            *error_models (BaseErrorModel): error models to be aggregated
            reduction (BaseReduction): reduction method used to aggregate uncertainty scores.
                Must output 1 component. Default FactorAnalysis().
    
    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000)
    model.fit(train["Drug"], train["Y"])
    error_model = oce.AggregateErrorModel(error_models = [oce.TargetDistDC(), oce.TrainDistDC()])
    error_model.build(model, train["Drug"], train["Y"])
    error_model.fit(valid["Drug"], valid["Y"])
    error_model.score(test["Drug"])
    ------------------------------
    """

    @log_arguments
    def __init__(
        self, 
        *error_models: BaseErrorModel, 
        reduction: BaseReduction = FactorAnalysis(n_components = 1),
        log=True,
        **kwargs
    ):
        self.error_models = list(error_models)
        self.reduction = reduction
        super().__init__(log=False, **kwargs)

    def _build(self, **kwargs):
        for error_model in self.error_models:
            error_model.build(self.model, self.X_train, self.y_train)

    def fit(self, X: Union[pd.DataFrame, np.ndarray, list, pd.Series], y: Union[np.ndarray, list, pd.Series], **kwargs):
        """Fits confidence scores to an external dataset

        Args:
            X (array-like): features, smiles
            y (array-like): true values
        
        Returns:
            plotly figure of fitted model against validation dataset
        """
        y_pred = np.array(self.model.predict(X)).flatten()
        scores = [error_model.calculate(X, y_pred) for error_model in self.error_models]
        scores = np.transpose(np.stack(scores))
        
        self.reduction.fit(scores)
        aggregate_scores = self.reduction.transform(scores).flatten()
        residuals = np.abs(np.array(y) - y_pred)
        self.fit_scores = scores

        return self._fit(residuals, aggregate_scores, **kwargs)

    def fit_cv(self, n_splits: int = 10, **kwargs):
        """Fits confidence scores to the training dataset via cross validation.

        Args:
            n_splits (int): Number of cross validation splits, default 5

        Returns:
            plotly figure of fitted model against validation dataset
        """
        from sklearn.model_selection import KFold

        self.X_train = np.array(self.X_train).flatten()

        residuals = None
        scores = None
        kf = KFold(n_splits=n_splits)

        split = 1
        for train_index, test_index in kf.split(self.X_train):
            print('evaluating split {} of {}'.format(split, n_splits))
            split += 1

            X_train, X_test = self.X_train[train_index], self.X_train[test_index]
            y_train, y_test = self.y_train[train_index], self.y_train[test_index]
            model = self.model.copy()
            model.fit(X_train, y_train)
            
            y_pred_test = np.array(model.predict(X_test)).flatten()
            pred_error = np.abs(y_test - y_pred_test)
            if residuals is None:
                residuals = pred_error
            else:
                residuals = np.concatenate((residuals, pred_error))

            for error_model in self.error_models:
                error_model.build(model, X_train, y_train)
            scores_fold = [error_model.calculate(X_test, y_pred_test) for error_model in self.error_models]
            scores_fold = np.transpose(np.stack(scores_fold))
            if scores is None:
                scores = scores_fold
            else:
                scores = np.concatenate((scores, scores_fold))
        
        self.reduction.fit(scores)
        aggregate_scores = self.reduction.transform(scores).flatten()
        self.fit_scores = scores

        return self._fit(residuals, aggregate_scores, **kwargs)

    def calculate(
        self, X: Union[pd.DataFrame, np.ndarray, list, pd.Series], y_pred: np.ndarray
    ) -> np.ndarray:
        """Computes aggregate error model score from inputs.

        Args:
            X: features, smiles
            y_pred: predicted values
        """
        scores = [error_model.calculate(X, y_pred) for error_model in self.error_models]
        scores = np.transpose(np.stack(scores))

        return self.reduction.transform(scores)
    
    def _save(self) -> dict:
        d = super()._save()
        if hasattr(self, "fit_scores"):
            d.update({"fit_scores": self.fit_scores})
        return d

    def _load(self, d) -> None:
        super()._load(d)
        if "fit_scores" in d.keys():
            self.fit_scores = d["fit_scores"]
            self.reduction.fit(self.fit_scores)


# Adapted from github.com/PatWalters/yamc under the MIT license
# which in turn states:
# A minor refactoring of code from
# https://github.com/Ryan-Rhys/The-Photoswitch-Dataset/blob/master/examples/gp_regression_on_molecules.ipynb

from unittest.mock import MagicMock
import numpy as np

import olorenchemengine as oce
from olorenchemengine.internal import *
from olorenchemengine.base_class import BaseEstimator, log_arguments, BaseSKLearnModel, BaseClass
from olorenchemengine.representations import BaseVecRepresentation

from sklearn.gaussian_process import GaussianProcessRegressor, GaussianProcessClassifier
from sklearn.gaussian_process.kernels import RBF, Matern, RationalQuadratic, ExpSineSquared, DotProduct, ConstantKernel, WhiteKernel, PairwiseKernel
from sklearn.gaussian_process.kernels import Hyperparameter, Kernel

class GaussianProcessRegressor(BaseEstimator):

    """Wrapper for sklearn GaussianProcessRegressor"""

    @log_arguments
    def __init__(self, *args, **kwargs):
        from sklearn.gaussian_process import GaussianProcessRegressor

        self.obj = GaussianProcessRegressor(*args, **kwargs)
        
class GaussianProcessClassifier(BaseEstimator):

    """Wrapper for sklearn GaussianProcessClassifier"""

    @log_arguments
    def __init__(self, *args, **kwargs):
        from sklearn.gaussian_process import GaussianProcessClassifier

        self.obj = GaussianProcessClassifier(*args, **kwargs)
class Tanimoto(Kernel):
    r"""Tanimoto kernel.
    
    Adapted from DotProduct kernel in sklearn.gaussian_process.kernels and from
    Ryan-Rhys Griffiths' implementation of the Tanimoto kernel in 
    https://towardsdatascience.com/gaussian-process-regression-on-molecules-in-gpflow-ee6fedab2130.
    """

    def __init__(self, sigma_0=1.0, sigma_0_bounds=(1e-5, 1e5)):
        self.sigma_0 = sigma_0
        self.sigma_0_bounds = sigma_0_bounds

    @property
    def hyperparameter_sigma_0(self):
        return Hyperparameter("sigma_0", "numeric", self.sigma_0_bounds)

    def bew(self, A, B):
        flatres = np.reshape(A, (-1, 1)) + np.reshape(B, (1, -1))
        return np.reshape(flatres, np.concatenate([A.shape, B.shape], 0))
    
    def __call__(self, X, Y=None, eval_gradient=False):
        """Return the kernel k(X, Y) and optionally its gradient.
        Parameters
        ----------
        X : ndarray of shape (n_samples_X, n_features)
            Left argument of the returned kernel k(X, Y)
        Y : ndarray of shape (n_samples_Y, n_features), default=None
            Right argument of the returned kernel k(X, Y). If None, k(X, X)
            if evaluated instead.
        eval_gradient : bool, default=False
            Determines whether the gradient with respect to the log of
            the kernel hyperparameter is computed.
            Only supported when Y is None.
        Returns
        -------
        K : ndarray of shape (n_samples_X, n_samples_Y)
            Kernel k(X, Y)
        K_gradient : ndarray of shape (n_samples_X, n_samples_X, n_dims),\
                optional
            The gradient of the kernel k(X, X) with respect to the log of the
            hyperparameter of the kernel. Only returned when `eval_gradient`
            is True.
        """
        X = np.atleast_2d(X)
        if Y is None:
            product = np.tensordot(X, X, axes=[[-1], [-1]])
            K = self.sigma_0**2 * product / (2*np.sum(X**2, axis=-1) - product)
        else:
            if eval_gradient:
                raise ValueError("Gradient can only be evaluated when Y is None.")
            product = np.tensordot(X, Y, axes=[[-1], [-1]])
            
            K = self.sigma_0**2 * product / (self.bew(np.sum(X**2, axis=-1), np.sum(Y**2, axis=-1)) - product)

        if eval_gradient:
            if not self.hyperparameter_sigma_0.fixed:
                return K, K[..., np.newaxis]*2*self.sigma_0**2
            else:
                return K, np.empty((X.shape[0], X.shape[0], 0))
        else:
            return K

    def diag(self, X):
        """Returns the diagonal of the kernel k(X, X).
        The result of this method is identical to np.diag(self(X)); however,
        it can be evaluated more efficiently since only the diagonal is
        evaluated.
        Parameters
        ----------
        X : ndarray of shape (n_samples_X, n_features)
            Left argument of the returned kernel k(X, Y).
        Returns
        -------
        K_diag : ndarray of shape (n_samples_X,)
            Diagonal of kernel k(X, X).
        """
        return np.diag(self(X))
    
    def is_stationary(self):
        """Returns whether the kernel is stationary."""
        return False

    def __repr__(self):
        return "{0}(sigma_0={1:.3g})".format(self.__class__.__name__, self.sigma_0)
    
class GaussianProcessModel(BaseSKLearnModel):
    """Gaussian process model using SciPy's implementation of Gaussian process models.
    
    Parameters:
        representation (BaseVecRepresentation): Representation to use.
        kernel (str): Kernel to use. Default is None which uses SciPy's default.
        kernel_params (dict): Parameters for the kernel. Default is {}, where no
            kernel parameters are passed and SciPy defaults are used.
        alpha (float): Value of alpha to use. Default is 1e-10.
        random_state (int): Random state to use. Default is None.
    """

    @log_arguments
    def __init__(
        self,
        representation: BaseVecRepresentation,
        kernel: str = None,
        kernel_params: dict = {},
        alpha: float = 1e-10,
        random_state: int = None,
        **kwargs
    ):
        if kernel == "RBF":
            kernel = RBF(**kernel_params)
        elif kernel == "Matern":
            kernel = Matern(**kernel_params)
        elif kernel == "DotProduct":
            kernel = DotProduct(**kernel_params)
        elif kernel == "WhiteKernel":
            kernel = WhiteKernel(**kernel_params)
        elif kernel == "RationalQuadratic":
            kernel = RationalQuadratic(**kernel_params)
        elif kernel == "ExpSineSquared":
            kernel = ExpSineSquared(**kernel_params)
        elif kernel == "PairwiseKernel":
            kernel = PairwiseKernel(**kernel_params)
        elif kernel == "ConstantKernel":
            kernel = ConstantKernel(**kernel_params)
        elif kernel == "Tanimoto":
            kernel = Tanimoto(**kernel_params)
        else:
            kernel = None
       
        regressor = GaussianProcessRegressor(
            kernel = kernel,
            alpha= alpha,
            random_state=random_state
        )
        classifier = GaussianProcessClassifier(
            kernel = kernel,
            random_state=random_state,
        )
        super().__init__(representation, regressor, classifier, log=False, **kwargs)

if package_available("gpflow"):
    from abc import abstractproperty
    import gpflow
    import tensorflow as tf
    import numpy as np
    from gpflow.mean_functions import Constant
    from sklearn.preprocessing import StandardScaler
    from gpflow.utilities import positive
    from gpflow.utilities.ops import broadcasting_elementwise

    import os
    from os import listdir
    import zipfile
    import io

    # A minor refactoring of code from
    # https://github.com/Ryan-Rhys/The-Photoswitch-Dataset/blob/master/examples/gp_regression_on_molecules.ipynb

    class Tanimoto(gpflow.kernels.Kernel):
        def __init__(self):
            super().__init__()
            # We constrain the value of the kernel variance to be positive when it's being optimised
            self.variance = gpflow.Parameter(1.0, transform=positive())

        def K(self, X, X2=None):
            """
            Compute the Tanimoto kernel matrix σ² * ((<x, y>) / (||x||^2 + ||y||^2 - <x, y>))

            :param X: N x D array
            :param X2: M x D array. If None, compute the N x N kernel matrix for X.
            :return: The kernel matrix of dimension N x M
            """
            if X2 is None:
                X2 = X

            Xs = tf.reduce_sum(tf.square(X), axis=-1)  # Squared L2-norm of X
            X2s = tf.reduce_sum(tf.square(X2), axis=-1)  # Squared L2-norm of X2
            outer_product = tf.tensordot(X, X2, [[-1], [-1]])  # outer product of the matrices X and X2

            # Analogue of denominator in Tanimoto formula

            denominator = -outer_product + broadcasting_elementwise(tf.add, Xs, X2s)

            return self.variance * outer_product / denominator

        def K_diag(self, X):
            """
            Compute the diagonal of the N x N kernel matrix of X
            :param X: N x D array
            :return: N x 1 array
            """
            return tf.fill(tf.shape(X)[:-1], tf.squeeze(self.variance))
    import tensorflow_probability.python.bijectors as bijectors

    class GPFlow_Parameter(BaseClass):
        """A wrapper for GPFlow parameters that makes it possible to save a load
        gpflow.Parameter classes."""

        @log_arguments
        def __init__(self, log=True):
            pass
        
        def set(self, parameter: gpflow.Parameter):
            self.d = {
                "value": parameter.numpy(),
                "name": parameter.name,
                "transform": parameter.transform.__class__.__name__,
            }
            return self

        def get(self):
            return gpflow.Parameter(self.d["value"], 
                                    name=self.d["name"],
                                    transform=getattr(bijectors, self.d["transform"])())

        def _save(self):
            return self.d

        def _load(self, d: dict):
            self.d = d

    class GPFlowEstimator(BaseEstimator):
        """Base class for GPFlow estimators. This class is not meant to be used directly.
        Use one of the derived classes instead. Derived classes implement the
        abstract method get_GPFlowModel, which returns a GPFlow model consstructed
        from a passed X_train and y_train.
        
        Requires installing GPFlow and TensorFlow.
        """

        @log_arguments
        def __init__(self, maxiter = 100, objective = "training-loss",
                     inducing_points = float('inf'), setting="regression", log=True):
            self.maxiter = maxiter
            self.m = None
            self.objective = objective
            self.setting = setting
            self.inducing_points = inducing_points
            self.inducing_variable = None

        @abstractmethod
        def get_GPFlowModel(self, X_train, y_train):
            pass
        
        def get_objective(self):
            if self.objective == "log_marginal_likelihood":
                if self.setting == "regression":
                    return lambda: -self.m.log_marginal_likelihood()
                else:
                    raise ValueError("log_marginal_likelihood is only implemented for regression")
            else:
                return gpflow.models.training_loss_closure(self.m, (self.X_train.astype(np.float64), np.reshape(self.y_train, (-1,1))))                        
        
        def fit(self, X_train, y_train):
            self.X_train = X_train
            self.y_train = y_train
            
            opt = gpflow.optimizers.Scipy()

            if X_train.shape[0] <= self.inducing_points:
                self.m = self.get_GPFlowModel(X_train, y_train)
                opt.minimize(self.get_objective(), self.m.trainable_variables, 
                            options=dict(maxiter=self.maxiter))
            else:
                rng = np.random.default_rng(1234)
                self.inducing_variable = rng.choice(X_train.astype(np.float64), size=self.inducing_points, replace=False)
                self.m = self.get_GPFlowModel(X_train, y_train, inducing_variable = self.inducing_variable.astype(np.float64))
                opt.minimize(self.get_objective(), self.m.trainable_variables, 
                            options=dict(maxiter=self.maxiter))
            tf.keras.backend.clear_session()

        def predict(self, X_test):
            y_pred, y_var = self.m.predict_f(X_test.astype(np.float64))
            return y_pred.numpy().flatten()

        def _save(self):
            parameter_dict_org = gpflow.utilities.parameter_dict(self.m)
            parameter_dict_new = {k: GPFlow_Parameter().set(v) for k, v in parameter_dict_org.items()}
            d = {"parameter_dict": {k: v._save() for k, v in parameter_dict_new.items()},
                 "x_train": self.X_train.tolist(),
                 "y_train": self.y_train.tolist(),
                 "inducing_variable": (self.inducing_variable.tolist() if self.inducing_variable is not None else None),
                 "inducing_points": self.inducing_points,}

            return d

        def _load(self, d):
            self.X_train = np.array(d["x_train"])
            self.y_train = np.array(d["y_train"])
            self.inducing_points = d["inducing_points"]
            self.inducing_variable = (np.array(d["inducing_variable"]) if d["inducing_variable"] is not None else None)
            self.m = self.get_GPFlowModel(self.X_train, self.y_train, inducing_variable = self.inducing_variable)
            parameter_dict_new = {k: GPFlow_Parameter() for k, v in d["parameter_dict"].items()}
            for k, v in d["parameter_dict"].items():
                parameter_dict_new[k]._load(v)
            parameter_dict_new = {k: v.get() for k, v in parameter_dict_new.items()}
            gpflow.utilities.multiple_assign(self.m, parameter_dict_new)

    class TanimotoGPRegressor(GPFlowEstimator):
        """
        Estimator for a GP regressor using the Tanimoto kernel.
        
        Adapted from Pat Walters github.com/PatWalters/yamc under the MIT license which in turn states:
        A minor refactoring of code from Ryan Rhys-Griffiths
        https://github.com/Ryan-Rhys/The-Photoswitch-Dataset/blob/master/examples/gp_regression_on_molecules.ipynb.
        """
        def get_GPFlowModel(self, X_train, y_train, inducing_variable = None):
            k = Tanimoto()
            if inducing_variable is None:
                m = gpflow.models.GPR(data=(X_train.astype(np.float64), np.reshape(y_train, (-1,1))),
                                    mean_function=Constant(np.mean(y_train)),
                                    kernel=k,
                                    noise_variance=1)
            else:
                m = gpflow.models.SGPR(data=(X_train.astype(np.float64), np.reshape(y_train, (-1,1))),
                                    mean_function=Constant(np.mean(y_train)),
                                    kernel = k,
                                    noise_variance = 1,
                                    inducing_variable = inducing_variable)
            return m

    class TanimotoGPClassifier(GPFlowEstimator):
        """
        Estimator for a GP classifier using the Tanimoto kernel.
        Requires installing GPFlow and TensorFlow.
        
        Adapted from Pat Walters github.com/PatWalters/yamc under the MIT license which in turn states:
        A minor refactoring of code from Ryan Rhys-Griffiths
        https://github.com/Ryan-Rhys/The-Photoswitch-Dataset/blob/master/examples/gp_regression_on_molecules.ipynb.
        """

        def get_GPFlowModel(self, X_train, y_train, inducing_variable = None):
            k = Tanimoto()
            if inducing_variable is None:
                m = gpflow.models.VGP(data=(X_train.astype(np.float64), np.reshape(y_train, (-1,1))),
                                   mean_function=Constant(np.mean(y_train)),
                                   kernel=k,
                                   likelihood=gpflow.likelihoods.Bernoulli())
            else:
                m = gpflow.models.SVGP(mean_function=Constant(np.mean(y_train)),
                                   kernel=k,
                                   likelihood=gpflow.likelihoods.Bernoulli(),
                                   inducing_variable=inducing_variable)
            return m

    class TanimotoGPModel(BaseSKLearnModel):
        """
        Gaussian Process model using the Tanimoto kernel implemented with GPFlow. 
        Requires installing GPFlow and TensorFlow.
        
        Adapted from Pat Walters github.com/PatWalters/yamc under the MIT license which in turn states:
        A minor refactoring of code from Ryan Rhys-Griffiths
        https://github.com/Ryan-Rhys/The-Photoswitch-Dataset/blob/master/examples/gp_regression_on_molecules.ipynb.
        
        Parameters:
            representation (BaseVecRepresentation): The representation to use."""

        @log_arguments
        def __init__(self, representation: BaseVecRepresentation, inducing_points = float('inf'),
                     objective = "training-loss",log=True, **kwargs):
            classifier = TanimotoGPClassifier(objective=objective, inducing_points = inducing_points, setting="classification")
            regressor = TanimotoGPRegressor(objective=objective, inducing_points = inducing_points, setting="regression")
            super().__init__(representation, regressor, classifier, log=False, **kwargs)
